"""ClinicalTrials.gov RDBMS model definition."""
import re

from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Table, Text, ForeignKey

from ebel.manager.rdbms.models import object_as_dict


Base = declarative_base()

ctg_keyword_n2m = Table('clinical_trials_gov__keyword', Base.metadata,
                        Column('clinical_trials_gov_keyword_id', Integer, ForeignKey('clinical_trials_gov_keyword.id'),
                               index=True),
                        Column('clinical_trials_gov_id', Integer,
                               ForeignKey('clinical_trials_gov.id'), index=True)
                        )

ctg_condition_n2m = Table('clinical_trials_gov__condition', Base.metadata,
                          Column('clinical_trials_gov_condition_id', Integer,
                                 ForeignKey('clinical_trials_gov_condition.id'),
                                 index=True),
                          Column('clinical_trials_gov_id', Integer,
                                 ForeignKey('clinical_trials_gov.id'), index=True)
                          )

ctg_mesh_term_n2m = Table('clinical_trials_gov__mesh_term', Base.metadata,
                          Column('clinical_trials_gov_mesh_term_id', Integer,
                                 ForeignKey('clinical_trials_gov_mesh_term.id'),
                                 index=True),
                          Column('clinical_trials_gov_id', Integer,
                                 ForeignKey('clinical_trials_gov.id'), index=True)
                          )

ctg_intervention_n2m = Table('clinical_trials_gov__intervention', Base.metadata,
                             Column('clinical_trials_gov_intervention_id', Integer,
                                    ForeignKey('clinical_trials_gov_intervention.id'),
                                    index=True),
                             Column('clinical_trials_gov_id', Integer,
                                    ForeignKey('clinical_trials_gov.id'), index=True)
                             )


class ClinicalTrialGov(Base):
    """Class definition for the clinical_trials_gov table."""

    __tablename__ = 'clinical_trials_gov'

    id = Column(Integer, primary_key=True)
    nct_id = Column(String(100), index=True)
    org_study_id = Column(Text)
    brief_title = Column(Text)
    official_title = Column(Text)
    is_fda_regulated_drug = Column(Text)
    brief_summary = Column(Text)
    detailed_description = Column(Text)
    overall_status = Column(Text)
    start_date = Column(Text)
    completion_date = Column(Text)
    phase = Column(Text)
    study_type = Column(Text)
    study_design_intervention_model = Column(Text)
    study_design_primary_purpose = Column(Text)
    study_design_masking = Column(Text)
    # primary_outcomes
    # secondary_outcomes
    patient_data_sharing_ipd = Column(Text)
    patient_data_ipd_description = Column(Text)

    keywords = relationship(
        "Keyword",
        secondary=ctg_keyword_n2m,
        back_populates="trials",
        cascade="save-update"
    )

    conditions = relationship(
        "Condition",
        secondary=ctg_condition_n2m,
        back_populates="trials",
        cascade="save-update"
    )

    mesh_terms = relationship(
        "MeshTerm",
        secondary=ctg_mesh_term_n2m,
        back_populates="trials",
        cascade="save-update"
    )

    interventions = relationship(
        "Intervention",
        secondary=ctg_intervention_n2m,
        back_populates="trials",
        cascade="save-update"
    )

    def as_dict(self):
        """Convert object values to dictionary."""
        basic_dict = object_as_dict(self)
        basic_dict['brief_summary'] = re.sub(r'\r?\n\s*', ' ', basic_dict['brief_summary']).strip()
        basic_dict.update({'keywords': [x.keyword for x in self.keywords]})
        basic_dict.update({'conditions': [x.condition for x in self.conditions]})
        basic_dict.update({'mesh_terms': [x.mesh_term for x in self.mesh_terms]})
        basic_dict.update({'interventions': [
            {'intervention_type': x.intervention_type,
             'intervention_name': x.intervention_name} for x in self.interventions]})
        return basic_dict


class Keyword(Base):
    """Class definition for the clinical_trials_gov_keyword table."""

    __tablename__ = 'clinical_trials_gov_keyword'
    id = Column(Integer, primary_key=True)
    keyword = Column(String(255, collation='utf8_bin'), index=True)
    trials = relationship(
        "ClinicalTrialGov",
        secondary=ctg_keyword_n2m,
        back_populates="keywords")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'keyword': self.keyword,
            'nct_ids': [x.nct_id for x in self.trials]
        }


class Condition(Base):
    """Class definition for the clinical_trials_gov_condition table."""

    __tablename__ = 'clinical_trials_gov_condition'
    id = Column(Integer, primary_key=True)
    condition = Column(Text)
    trials = relationship(
        "ClinicalTrialGov",
        secondary=ctg_condition_n2m,
        back_populates="conditions")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'condition': self.condition,
            'nct_ids': [x.nct_id for x in self.trials]
        }


class MeshTerm(Base):
    """Class definition for the clinical_trials_gov_mesh_term table."""

    __tablename__ = 'clinical_trials_gov_mesh_term'
    id = Column(Integer, primary_key=True)
    mesh_term = Column(String(100), unique=True)
    trials = relationship(
        "ClinicalTrialGov",
        secondary=ctg_mesh_term_n2m,
        back_populates="mesh_terms")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'mesh_term': self.mesh_term,
            'number_of_trials': len(self.trials),
            'nct_ids': [x.nct_id for x in self.trials]
        }


class Intervention(Base):
    """Class definition for the clinical_trials_gov_intervention table."""

    __tablename__ = 'clinical_trials_gov_intervention'
    id = Column(Integer, primary_key=True)
    intervention_type = Column(String(100, collation='utf8_bin'), index=True)
    intervention_name = Column(String(255, collation='utf8_bin'), index=True)
    trials = relationship(
        "ClinicalTrialGov",
        secondary=ctg_intervention_n2m,
        back_populates="interventions")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'intervention_type': self.intervention_type,
            'intervention_name': self.intervention_name,
            'nct_ids': [x.nct_id for x in self.trials]
        }



"""Pathway Commons API methods."""

from flask import request
from sqlalchemy import or_

from ebel.web.api import RDBMS
from ebel.manager.rdbms.models.pathway_commons import PathwayCommons, PathwayName, pathway_commons__pathway_name, Pmid
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result, _get_paginated_query_result, \
    _get_terms_from_model_starts_with


def get_pathway_commons():
    """Get paginated list of Pathway Commons entries."""
    return _get_data(PathwayCommons)


def get_by_gene_symbol():
    """Get entries by gene symbol."""
    gene_symbol = request.args.get('gene_symbol')
    query = RDBMS.get_session().query(PathwayCommons).filter(
        or_(PathwayCommons.participant_a == gene_symbol, PathwayCommons.participant_b == gene_symbol))
    return _get_paginated_query_result(query)


def get_by_pathway_name():
    """Get entries by pathway name."""
    pathway_name = request.args.get('pathway_name')
    query = RDBMS.get_session().query(PathwayCommons) \
        .join(pathway_commons__pathway_name) \
        .join(PathwayName) \
        .filter_by(name=pathway_name)
    return _get_paginated_query_result(query)


def get_by_pmid():
    """Get entries by PMID."""
    pmid = request.args.get('pmid')
    query = RDBMS.get_session().query(PathwayCommons) \
        .join(Pmid) \
        .filter_by(pmid=pmid)
    return _get_paginated_query_result(query)


def get_pathway_name_starts_with():
    """Get entries where pathway starts with given value."""
    return _get_terms_from_model_starts_with('pathway_name', PathwayName.name)


def get_ebel_relation():
    """Get Pathway Commons related eBEL relations."""
    conf = {
        'interactor_a_rid': "out.rid.asString()",
        'interactor_a_namespace': "out.namespace",
        'interactor_a_name': "out.name",
        'edge_rid': "rid.asString()",
        'relation_type': "type",
        'sources': "sources",
        'pmids': "pmids",
        'pathways': "pathways.name",
        'interactor_b_rid': "in.rid.asString()",
        'interactor_b_namespace': "in.namespace",
        'interactor_b_name': "in.name",
    }
    sql = "SELECT "
    sql += ', '.join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_action_pc"

    ra = request.args
    wheres = []

    source = ra.get('source')
    if source:
        wheres.append(f'"{source}" in sources')

    pmid = ra.get('pmid')
    if source:
        wheres.append(f'{pmid} in pmids')

    pathway = ra.get('pathway')
    if pathway:
        wheres.append(f'"{pathway}" in pathways.name')

    paras = {k: request.args[k] for k in request.args if k in conf}
    if paras:
        wheres += [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]

    if wheres:
        sql += " WHERE " + ' AND '.join(wheres)

    return _get_paginated_ebel_query_result(sql)



"""API server string constants."""
select_bel_columns_list = [
    'out.name as subject_name',
    'out.label as subject_label',
    'out.namespace as subject_namespace',
    'out.bel as subject_bel',
    'out.involved_genes as subject_involved_genes',
    'out.involved_other as subject_involved_other',
    'out.@class as subject_class',
    'out.@rid.asString() as subject_id',
    '@class as relation',
    'pmid',
    '@rid.asString() as edge_id',
    'evidence',
    'citation.last_author as last_author',
    'citation.title as title',
    'citation.pub_date as publication_date',
    'annotation',
    'in.name as object_name',
    'in.label as object_label',
    'in.namespace as object_namespace',
    'in.bel as object_bel',
    'in.involved_genes as object_involved_genes',
    'in.involved_other as object_involved_other',
    'in.@class as object_class',
    'in.@rid.asString() as object_id'
]

select_bel_columns = ', '.join(select_bel_columns_list)

match_bel_columns_list = [
    's.bel as subject_bel',
    's.involved_genes as subject_involved_genes',
    's.involved_other as subject_involved_other',
    's.@class as subject_class',
    's.@rid.asString() as subject_id',
    's.name as subject_name',
    's.label as subject_label',
    's.namespace as subject_namespace',
    'r.@class as relation',
    'r.pmid as pmid',
    'r.@rid.asString() as edge_id',
    'r.evidence as evidence',
    'r.citation.last_author as last_author',
    'r.citation.title as title',
    'r.citation.pub_date as publication_date',
    'r.annotation as annotation',
    'o.bel as object_bel',
    'o.involved_genes as object_involved_genes',
    'o.involved_other as object_involved_other',
    'o.@class as object_class',
    'o.@rid.asString() as object_id',
    'o.name as object_name',
    'o.label as object_label',
    'o.namespace as object_namespace',
]

match_bel_columns = ', '.join(match_bel_columns_list)



"""Command line interface to e(BE:L)."""

import re
import sys
import logging
import random
import string

from collections import namedtuple
from getpass import getpass
from urllib.parse import quote

import click
import pymysql

import ebel.database
from ebel import Bel, web
from ebel.tools import get_config_as_dict
from ebel.manager.orientdb.odb_meta import Graph
from ebel.manager.orientdb.constants import DRUGBANK
from ebel.constants import TerminalFormatting as TF
from ebel.validate import validate_bel_file, repair_bel_file

logger = logging.getLogger(__name__)


@click.group(help="BEL test framework Command Line Utilities on {}".format(sys.executable))
@click.version_option()
def main():
    """Entry method."""
    pass


# TODO(@ceb): Implement by default "keep database" and new database as option #prio1
# TODO(@ceb): SQLAlchemy connection string #prio1
@main.command()
@click.argument('bel_script_path')
@click.option('-f', '--force_new_db', is_flag=True, default=False, help='force recreation of new database')
@click.option('-l', '--line_by_line', is_flag=True, default=False, help='check script line by line')
@click.option('-r', '--reports', default=None,
              help='path(s) to report file(s) seperated by comma with suffix (.md, .txt, .csv, .tsv, .json, .html)')
@click.option('-v', '--bel_version', default="2_1", help='stores a report in a file')
@click.option('-t', '--tree', is_flag=True, default=False, help='shows tree')
# TODO: Implement new Cytoscape export file
# @click.option('-c', '--cytoscape', is_flag=True, default=False, help='creates cytoscape file')
@click.option('-s', '--sqlalchemy_connection_str', default=None, help="SQLAlchmy connection string")
@click.option('-j', '--json_file', is_flag=True, help="Create json file")
def validate(bel_script_path: str, force_new_db: bool, line_by_line: bool, reports: str,
             bel_version: str, tree: bool, sqlalchemy_connection_str: str,
             json_file: bool):
    """Validate a BEL file using the defined grammar."""
    validate_bel_file(bel_script_path=bel_script_path, force_new_db=force_new_db, line_by_line=line_by_line,
                      reports=reports, bel_version=bel_version, tree=tree,
                      sqlalchemy_connection_str=sqlalchemy_connection_str, json_file=json_file)


@main.command()
@click.argument('bel_script_path')
@click.option('-n', '--new_file_path', default=None,
              help='Path to write repaired file to. If none passed, will overwrite original file.')
def repair(bel_script_path: str, new_file_path: str):
    repair_bel_file(bel_script_path, new_file_path)


@main.command()
@click.argument('json_file_path')
@click.option('-e', '--extend', is_flag=True, default=True, help='Flag to disable "extension" during import')
@click.option('-g', '--p2g', is_flag=True, default=True, help='Flag to disable "protein2gene" during import')
@click.option('-s', '--skip_drugbank', is_flag=True, default=False, help='Flag to disable DrugBank')
@click.option('-i', '--include_subfolders', is_flag=True, default=False, help='Flag to enable directory walking')
@click.option('--drugbank_user', help='Valid username for DrugBank')
@click.option('--drugbank_password', help='Valid username for DrugBank')
@click.option('-n', '--odb_name', help='OrientDB database name')
@click.option('-u', '--odb_user', help='OrientDB user (with admin rights)')
@click.option('-p', '--odb_password', help='OrientDB user (with admin rights) password')
@click.option('-h', '--odb_server', help='OrientDB server name or URI')
@click.option('-o', '--odb_port', help='OrientDB server port')
@click.option('--odb_user_reader', help=' OrientDB user with only read rights')
@click.option('--odb_user_reader_password', help='OrientDB user with only read rights password')
@click.option('--odb_root_password', help='OrientDB root user password (only during database setup)')
@click.option('--kegg_species', help='KEGG species')
@click.option('--sqlalchemy_connection_string', help='schema is user:password@server/database')
@click.option('--snp_related_traits', help='key of SNP related traits in GWAS catalog')
@click.option('--drugbank_user', help='Drugbank user')
@click.option('--drugbank_password', help='DrugBank password')
def import_json(json_file_path: str,
                extend: bool,
                p2g: bool,
                skip_drugbank: bool,
                drugbank_user: str,
                drugbank_password: str,
                include_subfolders: bool,
                odb_name: str,
                odb_user: str,
                odb_password: str,
                odb_server: str,
                odb_port: str,
                odb_user_reader: str,
                odb_user_reader_password: str,
                odb_root_password: str,
                kegg_species: str,
                sqlalchemy_connection_string: str,
                snp_related_traits: str):
    """Import JSON into OrientDB.

    Parameters
    ----------
    json_file_path : str
        Path to a BEL JSON file, list of BEL JSON files (separated by ","), or a directory containing BEL JSON files.
    extend : bool
        Boolean flag to disable network extension by e(BE:L)
    p2g : bool
        Boolean flag to disable adding genes and RNA to proteins in network
    skip_drugbank : bool
        Boolean flag to disable DrugBank enrichment. Useful if you do not have valid DrugBank credentials
    drugbank_user: str (optional)
        DrugBank user name.
    drugbank_password: str (optional)
        DrugBank password.
    include_subfolders: bool (optional)
        Boolean flag to enable walking through subdirectories during import.
    odb_name: str
        OrientDB database name
    odb_user:
        OrientDB user (with admin rights)
    odb_password:
        OrientDB user (with admin rights) password
    odb_server:
        OrientDB server name or URI
    odb_port:
        OrientDB server port
    odb_user_reader:
        OrientDB user with only read rights
    odb_user_reader_password:
        OrientDB user with only read rights password
    odb_root_password:
        OrientDB root user password (only first database setup needed)
    kegg_species:
        KEGG species 3-4 letter code https://www.genome.jp/kegg/catalog/org_list4.html
    sqlalchemy_connection_string:
        SQL Alchemy connection string schema: user:passwd@server/database
    snp_related_traits:
        SNP related traits

    Returns
    -------
    type
        returns True if imported.
    """
    print("Start import JSON BEL file(s).")
    # if one of the parameters is not None it will overwrite the default values from the configfile
    Graph.set_configuration(name=odb_name,
                            user=odb_user,
                            password=odb_password,
                            server=odb_server,
                            port=odb_port,
                            user_reader=odb_user_reader,
                            user_reader_password=odb_user_reader_password,
                            root_password=odb_root_password,
                            kegg_species=kegg_species,
                            sqlalchemy_connection_string=sqlalchemy_connection_string,
                            snp_related_traits=snp_related_traits,
                            drugbank_user=drugbank_user,
                            drugbank_password=drugbank_password)

    bel = Bel()
    if "," in json_file_path:
        json_file_path = json_file_path.split()

    bel.import_json(input_path=json_file_path,
                    extend_graph=extend,
                    update_from_protein2gene=p2g,
                    skip_drugbank=skip_drugbank,
                    drugbank_user=drugbank_user,
                    drugbank_password=drugbank_password,
                    include_subfolders=include_subfolders)


@main.command()
@click.option('-s', '--skip', default=[], help='Comma-separated list of databases to skip during enrichment')
@click.option('-i', '--include', default=[], help='Comma-separated list of databases to include during enrichment')
@click.option('--skip_drugbank', is_flag=True, default=False, help='Flag to disable DrugBank')
@click.option('--drugbank_user', default=None, help='Valid username for DrugBank')
@click.option('--drugbank_password', default=None, help='Valid username for DrugBank')
@click.option('-n', '--odb_name', default=None, help='OrientDB database name')
@click.option('-u', '--odb_user', default=None, help='OrientDB user (with admin rights)')
@click.option('-p', '--odb_password', default=None, help='OrientDB user (with admin rights) password')
@click.option('-h', '--odb_server', default=None, help='OrientDB server name or URI')
@click.option('-o', '--odb_port', default=None, help='OrientDB server port')
@click.option('--odb_user_reader', default=None, help=' OrientDB user with only read rights')
@click.option('--odb_user_reader_password', default=None, help='OrientDB user with only read rights password')
@click.option('--odb_root_password', default=None, help='OrientDB root user password (only during database setup)')
@click.option('--kegg_species', default='hsa,rno,mmu', help='KEGG species')
@click.option('--sqlalchemy_connection_string', help='schema is user:password@server/database')
@click.option('--snp_related_traits', help='key of SNP related traits in GWAS catalog and ClinVar')
def enrich(skip: str,
           include: str,
           skip_drugbank: bool,
           drugbank_user: str,
           drugbank_password: str,
           odb_name: str,
           odb_user: str,
           odb_password: str,
           odb_server: str,
           odb_port: str,
           odb_user_reader: str,
           odb_user_reader_password: str,
           odb_root_password: str,
           kegg_species: str,
           sqlalchemy_connection_string: str,
           snp_related_traits: str):
    """Trigger the enrichment step for a database.

    Parameters
    ----------
    skip : str
        Comma-separated list of databases to skip during enrichment.
    include : str
        Comma-separated list of databases to include during enrichment.
    skip_drugbank : bool
        Boolean flag to disable DrugBank enrichment. Useful if you do not have valid DrugBank credentials
    drugbank_user: str (optional)
        DrugBank user name.
    drugbank_password: str (optional)
        DrugBank password.
    odb_name: str
        OrientDB database name
    odb_user: str
        OrientDB user (with admin rights)
    odb_password: str
        OrientDB user (with admin rights) password
    odb_server: str
        OrientDB server name or URI
    odb_port: int
        OrientDB server port
    odb_user_reader: str
        OrientDB user with only read rights
    odb_user_reader_password:
        OrientDB user with only read rights password
    odb_root_password: str
        OrientDB root user password (only first database setup needed)
    kegg_species: str
        KEGG species 3-4 letter code https://www.genome.jp/kegg/catalog/org_list4.html
    sqlalchemy_connection_string:
        SQL Alchemy connection string schema: user:passwd@server/database
    snp_related_traits: str
        SNP related traits
    """
    Graph.set_configuration(name=odb_name,
                            user=odb_user,
                            password=odb_password,
                            server=odb_server,
                            port=odb_port,
                            user_reader=odb_user_reader,
                            user_reader_password=odb_user_reader_password,
                            root_password=odb_root_password,
                            kegg_species=kegg_species,
                            sqlalchemy_connection_string=sqlalchemy_connection_string,
                            snp_related_traits=snp_related_traits,
                            drugbank_user=drugbank_user,
                            drugbank_password=drugbank_password)

    bel = Bel()

    if not skip_drugbank:
        bel.drugbank.get_user_passwd(drugbank_user=drugbank_user, drugbank_password=drugbank_password)

    skip = skip.split(",") if isinstance(skip, str) else skip
    include = include.split(",") if isinstance(include, str) else include

    if skip_drugbank and DRUGBANK not in skip:
        skip.append(DRUGBANK)

    bel.enrich_network(skip=skip, include=include)


@main.command()
@click.argument('connection')
def set_connection(connection):
    """Set the SQLAlchemy connection string.

    :param connection: An existing RDBMS connection.
    """
    ebel.database.set_connection(connection)


@main.command()
@click.option('-h', '--host', default='localhost', help="MySQL server")
@click.option('-u', '--user', default='ebel_user', help="MySQL username")
@click.option('-p', '--password', default='ebel_passwd', help="MySQL password")
@click.option('-d', '--database', default='ebel', help="MySQL database name")
@click.option('-i', '--interactive', is_flag=True, default=False, help="Enable interactive mode")
def set_mysql(host: str, user: str, password: str, database: str, interactive: bool):
    """Set the SQLAlchemy connection string with MySQL settings."""
    if interactive:
        host, user, password, db = ebel.database.set_mysql_interactive()

    ebel.database.set_mysql_connection(
        host=host,
        user=user,
        password=password,
        db=database
    )


@main.command()
def settings():
    """Interactive method to create a configuration file."""
    old_configs = get_config_as_dict()
    configs = {}
    allowed_passwd_chars = string.ascii_letters

    print(f"\n{TF.TITLE}e(BE:L) Configuration WIZARD{TF.RESET}")
    print("\nThe following questionnaire will set all configurations for you.\n")
    print("Before we start: Make sure\n\t1. OrientDB and\n\t2. MySQL/MariaDB\n are running and "
          "you have the root password for both,\n if databases and users not already exists.\n")
    print(f"Default {TF.DEFAULT_VALUE}[values]{TF.RESET} are written in square brackets and "
          f"can be confirmed by RETURN.\n")

    print(f"{TF.Format.UNDERLINED}Installation references{TF.RESET}")
    print(f"\t OrientDB: {TF.Fore.BLUE}https://orientdb.org/docs/3.1.x/fiveminute/java.html{TF.RESET}")
    print(f"\t MySQL: {TF.Fore.BLUE}"
          f"https://dev.mysql.com/doc/mysql-getting-started/en/#mysql-getting-started-installing{TF.RESET}")
    print(f"\t MariaDB: {TF.Fore.BLUE}https://mariadb.com/kb/en/getting-installing-and-upgrading-mariadb/{TF.RESET}")

    print(f"\n{TF.HEADER}Graph database (OrientDB) settings{TF.RESET}")

    Configuration = namedtuple('Configuration', ['name',
                                                 'question',
                                                 'default',
                                                 'validation_regex',
                                                 'is_password',
                                                 'required'])

    old_configs_odb = old_configs.get('DEFAULT_ODB', {})
    orientdb_configs = {
        'name': (f"OrientDB database name (created if not exists)",
                 old_configs_odb.get('name', 'ebel'), r'^[A-Za-z]\w{2,}$', False, True),
        'user': ("OrientDB user (admin) name (created if not exists)",
                 old_configs_odb.get('user', 'ebel_user'), r'^[A-Za-z]\w{2,}$', False, True),
        'password': ("OrientDB user (admin) password (created if not exists)",
                     ''.join(random.sample(allowed_passwd_chars, 12)), None, True, True),
        'server': ("OrientDB server",
                   old_configs_odb.get('server', 'localhost'), None, False, True),
        'port': ("OrientDB port",
                 old_configs_odb.get('port', '2424'), r'^\d+$', False, True),
        'user_reader': ("OrientDB user (reader) name (created if not exists).",
                        old_configs_odb.get('user_reader', 'ebel_reader'), r'^[A-Za-z]\w{2,}$', False, False),
        'user_reader_password': ("OrientDB user (reader) password (created if not exists).",
                                 ''.join(random.sample(allowed_passwd_chars, 12)), None, True, True),
    }

    for param_name, options in orientdb_configs.items():
        conf = Configuration(param_name, *options)
        input_method = getpass if conf.is_password else input
        invalid_value = True
        while invalid_value:
            if conf.default:
                question = f"{TF.QUESTION}{conf.question}{TF.RESET} {TF.DEFAULT_VALUE}[{conf.default}]{TF.RESET}"
                configs[conf.name] = input_method(f"\n{question} ?\n").strip() or conf.default
            else:
                configs[conf.name] = input_method(f"\n{TF.QUESTION}{conf.question}{TF.RESET}\n").strip()
            if conf.validation_regex and conf.required:
                invalid_value = not bool(re.search(conf.validation_regex, configs[conf.name]))
                if invalid_value:
                    print("!!!>>> WARNING <<<!!!\nThe value you type in is not valid. Please type in again.")
            else:
                invalid_value = False

    ebel.database.get_orientdb_client(
        server=configs['server'],
        port=int(configs['port']),
        name=configs['name'],
        user=configs['user'],
        password=configs['password'],
        root_password=None,
        user_reader=configs['user_reader'],
        user_reader_password=configs['user_reader_password']
    )

    print(f"\n{TF.HEADER}KEGG settings{TF.RESET}")

    kegg_question = f"{TF.QUESTION}KEGG species as 3-4 letter code comma separated.{TF.RESET}\n" \
                    "(see here for the letter code: " \
                    f"{TF.Fore.BLUE}https://www.genome.jp/kegg/catalog/org_list4.html{TF.RESET} )\n" \
                    f"{TF.DEFAULT_VALUE}[hsa,rno,mmu]{TF.RESET}\n"
    configs['kegg_species'] = input(kegg_question).strip() or 'hsa,rno,mmu'

    print(f"\n{TF.HEADER}MySQL/MariaDB settings{TF.RESET}")

    old_mysql = {}
    old_sa_con_str = old_configs.get('sqlalchemy_connection_string', '').strip()
    if old_sa_con_str:
        regex_con_str = r"^mysql\+pymysql://(?P<mysql_user>.*?):" \
                        r"(?P<mysql_passwd>.*?)@(?P<mysql_host>.*?)/(?P<mysql_db>.*)$"
        found_old_mysql = re.search(regex_con_str, old_sa_con_str)
        if found_old_mysql:
            old_mysql = found_old_mysql.groupdict()

    default_mysql_host = old_mysql.get('mysql_host') or 'localhost'
    mysql_host_question = f"{TF.QUESTION}MySQL/MariaDB sever name{TF.RESET} " \
                          f"{TF.DEFAULT_VALUE}[{default_mysql_host}]{TF.RESET}\n"
    mysql_host = input(mysql_host_question) or default_mysql_host
    default_mysql_user = old_mysql.get('mysql_user') or 'ebel'
    mysql_user_question = f"{TF.QUESTION}MySQL/MariaDB ebel user{TF.RESET} " \
                          f"{TF.DEFAULT_VALUE}[{default_mysql_user}]{TF.RESET}\n"
    mysql_user = input(mysql_user_question).strip() or default_mysql_user
    mysql_random_password = ''.join(random.sample(allowed_passwd_chars, 12))
    mysql_passed_question = f"{TF.QUESTION}MySQL/MariaDB ebel database password{TF.RESET} " \
                            f"{TF.DEFAULT_VALUE}[{mysql_random_password}]{TF.RESET}\n"
    mysql_passwd = getpass(mysql_passed_question).strip() or mysql_random_password
    default_mysql_db = old_mysql.get('mysql_db') or 'ebel'
    mysql_db_question = f"{TF.QUESTION}MySQL ebel database name{TF.RESET} " \
                        f"{TF.DEFAULT_VALUE}[{default_mysql_db}]{TF.RESET}\n"
    mysql_db = input(mysql_db_question).strip() or default_mysql_db
    configs['sqlalchemy_connection_string'] = f"{mysql_user}:{quote(mysql_passwd)}@" \
                                              f"{mysql_host}/{mysql_db}?charset=utf8mb4"
    try:
        pymysql.connect(host=mysql_host, user=mysql_user, password=mysql_passwd, db=mysql_db)
    except pymysql.err.OperationalError:
        mysql_root_passwd = getpass(f"{TF.QUESTION}MySQL root password (will be not stored) "
                                    f"to create database and user{TF.RESET}\n")
        print(mysql_host, 'root', mysql_root_passwd)
        cursor = pymysql.connect(host=mysql_host, user='root', password=mysql_root_passwd).cursor()
        db_exists = cursor.execute(f"show databases like %s", mysql_db)
        if not db_exists:
            cursor.execute(f"CREATE DATABASE {mysql_db} CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci")
        else:
            print(f"Database {mysql_db} alreay exists.")

        user_exists = cursor.execute(f"Select 1 from mysql.user where User=%s", mysql_user)
        if user_exists:
            print("Recreate user, because already exists")
            cursor.execute(f"DROP USER '{mysql_user}'@'%'")
            cursor.execute("FLUSH PRIVILEGES")
        sql = f"CREATE USER '{mysql_user}'@'%' IDENTIFIED BY '{mysql_passwd}'"
        print(sql)
        cursor.execute(sql)
        cursor.execute("FLUSH PRIVILEGES")

        privileges_exists = cursor.execute("Select 1 from mysql.db where User=%s and Db=%s", (mysql_user, mysql_db))
        if not privileges_exists:
            sql = f"GRANT ALL PRIVILEGES ON `{mysql_db}`.*  TO '{mysql_user}'@'%'"
            print(sql)
            cursor.execute(sql)
            cursor.execute("FLUSH PRIVILEGES")
        else:
            print(f"MySQL user {mysql_user} have alreay the rights for {mysql_db}.")

    print(f"\n{TF.HEADER}SNP related traits settings{TF.RESET}")

    default_snp_related_traits = old_configs.get('snp_related_traits') or 'Alzheimer,Parkinson'
    snp_related_traits_question = f"{TF.QUESTION}SNPs related to (separated by comma){TF.RESET} " \
                                  f"{TF.DEFAULT_VALUE}[{default_snp_related_traits}]{TF.RESET}\n"
    snp_related_traits = input(snp_related_traits_question).strip()
    configs['snp_related_traits'] = snp_related_traits or default_snp_related_traits

    print(f"\n{TF.HEADER}DrugBank settings{TF.RESET}")
    print("\nIf you have a valid DrugBank account (otherwise leave it blank):\n")
    drugbank_user = input(f"{TF.QUESTION}DrugBank user?{TF.RESET}\n").strip()
    if drugbank_user:
        configs['drugbank_user'] = drugbank_user
        configs['drugbank_password'] = getpass(f"{TF.QUESTION}DrugBank password?{TF.RESET}\n")

    Graph.set_configuration(**configs)


@main.command()
@click.option('-h', '--host', default='0.0.0.0', help='Server or host name')
@click.option('-p', '--port', default='5000', help='server port [5000]')
@click.option('-d', '--debug_mode', is_flag=True, default=False, help='debug mode')
@click.option('-o', '--open_browser', is_flag=True, default=False, help='open browser')
def serve(host, port, debug_mode, open_browser):
    """Start the API RESTful server."""
    web.app.run(host=host, port=port, debug_mode=debug_mode, open_browser=open_browser)


if __name__ == '__main__':
    main()



"""UniProt API methods."""
from flask import request

from ebel import Bel
from ebel.web.api import RDBMS
from ebel.manager.rdbms.models import uniprot
from ebel.web.api.ebel.v1 import _get_paginated_query_result, _get_terms_from_model_starts_with


def get_uniprot():
    """Get paginated list of UniProt entries."""
    q = RDBMS.get_session().query(uniprot.Uniprot)

    accession = request.args.get('accession')
    q = q.filter_by(accession=accession) if accession else q

    taxid = request.args.get('taxonomy_id')
    q = q.filter_by(taxid=taxid) if taxid else q

    gene_symbol = request.args.get('gene_symbol')
    if gene_symbol:
        q = q.join(uniprot.GeneSymbol).filter_by(symbol=gene_symbol)

    keyword = request.args.get('keyword')
    if keyword:
        q = q.join(uniprot.uniprot__uniprot_keyword) \
            .join(uniprot.Keyword).filter_by(keyword_name=keyword)

    xref_db = request.args.get('xref_db')
    xref_id = request.args.get('xref_id')
    if xref_db or xref_id:
        q = q.join(uniprot.uniprot__uniprot_xref).join(uniprot.Xref)
        q = q.filter_by(db=xref_db) if xref_db else q
        q = q.filter_by(identifier=xref_id) if xref_id else q

    subcellular_location = request.args.get('subcellular_location')
    if subcellular_location:
        q = q.join(uniprot.uniprot__uniprot_subcellular_location) \
            .join(uniprot.SubcellularLocation).filter_by(name=subcellular_location)

    return _get_paginated_query_result(q)


def get_keyword_starts_with():
    """Get entries where keyword starts with given value."""
    return _get_terms_from_model_starts_with('keyword', uniprot.Keyword.keyword_name)


def get_subcellular_location_starts_with():
    """Get entries where subcellular location starts with given value."""
    return _get_terms_from_model_starts_with('subcellular_location', uniprot.SubcellularLocation.name)


def get_gene_symbol_starts_with():
    """Get entries where symbol starts with given value."""
    return _get_terms_from_model_starts_with('gene_symbol', uniprot.GeneSymbol.symbol)


def get_gene_starts_with():
    """Get entries where gene starts with given value."""
    return _get_terms_from_model_starts_with('gene', uniprot.Gene.name)


def get_organism_starts_with():
    """Get entries where organism starts with given value."""
    return _get_terms_from_model_starts_with('organism', uniprot.Organism.scientific_name)


def get_function_starts_with():
    """Get entries where description starts with given value."""
    return _get_terms_from_model_starts_with('description', uniprot.Function.description)


def get_bel_node_uniprot():
    """Get UniProt related eBEL nodes."""
    b = Bel()
    conf = {
        'rid': "@rid.asString()",
        'name': "name",
        'namespace': "namespace",
        'bel': "bel",
        'uniprot_accession': "uniprot"
    }
    sql = "SELECT "
    sql += ', '.join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM protein WHERE pure = true"

    uniprot = request.args.get('accession')

    if uniprot:
        sql += f' AND uniprot = "{uniprot}"'
    else:
        return {'error': "uniprot_accession is required."}

    entries = b.query_get_dict(sql)
    if entries:
        result_dict = entries[0]
        result_dict['edges'] = {}
        for direction in ('in', 'out'):
            match = "match {class:protein, where:(uniprot='"
            match += uniprot + "' and pure=true)}." + direction + "E(bel_relation){as:e}"
            match += " return e.@rid, e.@class as relation"
            sql_edges = f"""Select relation, count(*)
                    from ({match}) group by relation order by count desc"""
            result_dict['edges'][direction] = b.query_get_dict(sql_edges)

        return result_dict



"""OFFSIDES API unit tests."""

from .conftest import format_response_data
from .constants import RESULTS, NUM_RESULTS, PAGE_SIZE


class TestOffsides:
    def test_get_offsides(self, client):
        response = client.get(
            'api/v1/offsides?drug_rxnorn_id=4024&drug_concept_name=ergoloid%20mesylates%2C%20USP&condition_meddra_id=10002034&condition_concept_name=Anaemia&a=6&b=126&c=21&d=1299&prr=2.85714&prr_error=0.45382&mean_reporting_frequency=0.0454545&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "a": 6,
          "b": 126,
          "c": 21,
          "condition_concept_name": "Anaemia",
          "condition_meddra_id": 10002034,
          "d": 1299,
          "drug_concept_name": "ergoloid mesylates, USP",
          "drug_rxnorn_id": "4024",
          "id": 1,
          "mean_reporting_frequency": 0.0454545,
          "prr": 2.85714,
          "prr_error": 0.45382
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""IntAct API methods."""
from sqlalchemy import or_
from flask.globals import request

from ebel.web.api import RDBMS
from ebel.manager.rdbms.models.intact import Intact
from ebel.manager.orientdb.odb_structure import intact_edges
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_query_result, _get_paginated_ebel_query_result


def get_intact():
    """Get generic IntAct entry."""
    return _get_data(Intact)


def get_by_uniprot():
    """Get IntAct entry by UniProt ID."""
    ua = request.args.get('uniprot_accession')
    if ua:
        a = Intact.int_a_uniprot_id
        b = Intact.int_a_uniprot_id
        query = RDBMS.get_session().query(Intact).filter(or_(a == ua, b == ua))
        return _get_paginated_query_result(query)


def get_ebel_relation():
    """Get IntAct related eBEL relations."""
    has_ppi_ia_edge = [x for x in intact_edges if x.name == 'has_ppi_ia'][0]
    conf = {x.prop_name: x.prop_name for x in has_ppi_ia_edge.props}
    conf.update({
        'relation_type': "@class",
        'edge_id': "@rid.asString()",
        'interactor_a_rid': "out.@rid.asString()",
        'interactor_a_name': "out.name",
        'interactor_a_namespace': "out.namespace",
        'interactor_a_bel': "out.bel",
        'interactor_b_rid': "in.@rid.asString()",
        'interactor_b_namespace': "in.namespace",
        'interactor_b_name': "in.name",
        'interactor_b_bel': "in.bel",
    })
    sql = "SELECT "
    sql += ', '.join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_ppi_ia"

    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}
    if paras:
        sql += " WHERE " + ' AND '.join([f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()])

    return _get_paginated_ebel_query_result(sql)



"""miRTarBase."""
import pandas as pd

from tqdm import tqdm
from typing import Dict
from pyorient import OrientDB

from ebel.tools import get_file_path
from ebel.manager.orientdb.constants import MIRTARBASE
from ebel.manager.orientdb import odb_meta, urls, odb_structure

from ebel.manager.rdbms.models import mirtarbase


class MirTarBase(odb_meta.Graph):
    """MirTarBase (from NCBI)."""

    def __init__(self, client: OrientDB = None):
        """Init MirTarBase (http://mirtarbase.mbc.nctu.edu.tw)."""
        self.client = client
        self.biodb_name = MIRTARBASE
        self.urls = {self.biodb_name: urls.MIRTARBASE}
        self.file_path = get_file_path(urls.MIRTARBASE, self.biodb_name)
        super().__init__(urls=self.urls,
                         tables_base=mirtarbase.Base,
                         edges=odb_structure.mirtarbase_edges,
                         biodb_name=self.biodb_name)

    def __len__(self) -> Dict[str, int]:
        return self.number_of_generics

    def __contains__(self, item) -> bool:
        # TODO: To be implemented
        return True

    def insert_data(self) -> Dict[str, int]:
        """Insert mirtarbase data into database."""
        df = pd.read_excel(self.file_path)
        df.columns = self._standardize_column_names(df.columns)
        df.index += 1
        df.index.rename('id', inplace=True)

        df.to_sql(mirtarbase.Mirtarbase.__tablename__, self.engine, if_exists='append', chunksize=10000)

        return {self.biodb_name: df.shape[0]}

    def update_interactions(self) -> int:
        """Update edges with mirtarbase metadata."""
        self.clear_edges()
        df_symbol_rid = self.get_pure_symbol_rid_df_in_bel_context(class_name='rna', namespace='HGNC')

        sql = f"""Select
                mi_rna,
                target_gene as symbol,
                support_type,
                references_pmid as pmid,
                experiments
            from
                {mirtarbase.Mirtarbase.__tablename__}
            where
                species_mi_rna='Homo sapiens' and
                species_target_gene='Homo sapiens' and
                support_type in ('Functional MTI', 'Non-Functional MTI')"""
        cols = ['mi_rna', 'symbol', 'support_type', 'pmid', 'experiments']
        df_mirtarbase = pd.DataFrame(self.engine.execute(sql).fetchall(), columns=cols)
        df_mirtarbase.experiments = df_mirtarbase.experiments.str.split('//')
        df_join = df_mirtarbase.set_index('symbol').join(df_symbol_rid.set_index('symbol'), how='inner')

        desc = f"Update {self.biodb_name.upper()} interactions"

        updated = 0
        for protein_rid, row in tqdm(df_join.set_index('rid').iterrows(), total=df_join.shape[0], desc=desc):
            mir_data = {'bel': f'm(MIRBASE:"{row.mi_rna}")',
                        'name': row.mi_rna,
                        'namespace': "MIRBASE",
                        'pure': True}
            mir_rid = self.get_create_rid('micro_rna', mir_data, check_for='bel')
            value_dict = {'support_type': row.support_type,
                          'pmid': row.pmid,
                          'experiments': row.experiments}
            self.create_edge('has_mirgene_target', mir_rid, str(protein_rid), value_dict=value_dict)
            updated += 1

        return updated



"""ClinVar."""
import logging
import pandas as pd

from tqdm import tqdm
from pyorient import OrientDB
from typing import Dict, List
from collections import namedtuple

from ebel.manager.orientdb.constants import CLINVAR
from ebel.manager.orientdb import odb_meta, urls, odb_structure
from ebel.tools import get_file_path, get_disease_trait_keywords_from_config

from ebel.manager.rdbms.models import clinvar


logger = logging.getLogger(__name__)

Snp = namedtuple('Snp', ('keyword',
                         'phenotype',
                         'rs_number',
                         'hgnc_id',
                         'chromosome',
                         'position',
                         'clinical_significance'))


class ClinVar(odb_meta.Graph):
    """ClinVar (from NCBI)."""

    def __init__(self, client: OrientDB = None):
        """Init ClinVar."""
        self.client = client
        self.biodb_name = CLINVAR
        self.urls = {self.biodb_name: urls.CLINVAR}
        self.file_path = get_file_path(urls.CLINVAR, self.biodb_name)
        super().__init__(nodes=odb_structure.clinvar_nodes,
                         edges=odb_structure.clinvar_edges,
                         indices=odb_structure.clinvar_indices,
                         urls=self.urls,
                         tables_base=clinvar.Base,
                         biodb_name=self.biodb_name)

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    def insert_data(self) -> Dict[str, int]:
        """Insert data."""
        inserted = {}
        self.recreate_tables()
        df = pd.read_csv(self.file_path, sep="\t", low_memory=False)
        self._standardize_dataframe(df)
        df.index += 1
        df.index.rename('id', inplace=True)
        df.drop(columns=['phenotype_ids', 'phenotype_list', 'other_ids']).to_sql(
            self.biodb_name, self.engine, if_exists='append', chunksize=10000)

        df_clinvar__phenotype = df['phenotype_list'].str.split(r'[|;]').explode().to_frame() \
            .reset_index().rename(columns={'phenotype_list': 'phenotype', 'id': 'clinvar_id'})
        df_clinvar__phenotype.index += 1
        df_phenotype = pd.DataFrame(df_clinvar__phenotype.phenotype.unique(), columns=['phenotype'])
        df_phenotype.index += 1
        df_phenotype.index.rename('id', inplace=True)
        df_phenotype.to_sql(clinvar.ClinvarPhenotype.__tablename__, self.engine, if_exists='append')
        inserted.update({clinvar.ClinvarPhenotype.__tablename__: df_phenotype.shape[0]})

        df_clinvar__phenotype = df_clinvar__phenotype.set_index('phenotype').join(
            df_phenotype.assign(clinvar_phenotype_id=df_phenotype.index).set_index('phenotype')
        ).reset_index().loc[:, ['clinvar_id', 'clinvar_phenotype_id']]

        df_clinvar__phenotype.index += 1
        df_clinvar__phenotype.index.rename('id', inplace=True)
        df_clinvar__phenotype.to_sql('clinvar__phenotype', self.engine, if_exists='append', index=False)
        inserted.update({clinvar.clinvar__clinvar_phenotype.__dict__['fullname']: df_clinvar__phenotype.shape[0]})

        clinvar_pheno_medgen = df['phenotype_ids'].str.split(r'[|,;]').explode().str.partition(':')[[0, 2]] \
            .rename(columns={0: 'db', 2: 'identifier'})
        clinvar_pheno_medgen['clinvar_id'] = clinvar_pheno_medgen.index
        clinvar_phenotype_medgen = clinvar_pheno_medgen.set_index('db').loc['MedGen'].reset_index().drop(
            columns=['db'])
        clinvar_phenotype_medgen.index += 1
        clinvar_phenotype_medgen.index.rename('id', inplace=True)
        clinvar_phenotype_medgen.to_sql(clinvar.ClinvarPhenotypeMedgen.__tablename__, self.engine, if_exists='append')
        inserted.update({clinvar.ClinvarPhenotypeMedgen.__tablename__: clinvar_phenotype_medgen.shape[0]})

        df_other_ids = df.other_ids.str.split(',').dropna().explode().str.partition(':')[[0, 2]].reset_index().rename(
            columns={0: 'db', 2: 'identifier', 'id': 'clinvar_id'}).rename_axis("id")
        df_other_ids.index += 1
        df_other_ids.to_sql(clinvar.ClinvarOtherIdentifier.__tablename__, self.engine, if_exists='append')
        inserted.update({clinvar.ClinvarOtherIdentifier.__tablename__: df_other_ids.shape[0]})
        return inserted

    def update_bel(self) -> int:
        """Delete and updates all ClinVar edges."""
        logger.info(f"Clear edges for {self.biodb_name.upper()}")
        self.clear_edges()
        logger.info("Clear nodes without edges")
        self.clear_nodes_with_no_edges()

        added_edges = self.update_interactions()
        logger.info(f"Successfully updated interactions for {self.biodb_name.upper()}")
        return added_edges

    def get_disease_snps_dict(self) -> Dict[str, List[Snp]]:
        """Get a dictionary {'disease':[snp,snp,... ]} by disease names."""
        disease_keywords = get_disease_trait_keywords_from_config()

        sql_temp = """Select
            '{keyword}',
            phenotype,
            rs_db_snp as rs_number,
            hgnc_id,
            chromosome,
            start as position,
            clinical_significance
                from clinvar c inner join
                clinvar__phenotype cp on (c.id=cp.clinvar_id) inner JOIN
                clinvar_phenotype p on (cp.clinvar_phenotype_id=p.id)
            where
                p.phenotype like '%%{keyword}%%'
                and rs_db_snp != -1"""

        results = dict()
        for kwd in disease_keywords:
            sql = sql_temp.format(keyword=kwd)
            rows = self.engine.execute(sql)
            results[kwd] = [Snp(*x) for x in rows.fetchall()]

        return results

    def update_interactions(self) -> int:
        """Create SNPs and edges using information from ClinVar."""
        # snp is upstream of a downstream gene
        snp_type = {'mapped': "mapped", 'downstream': "upstream", 'upstream': "downstream"}
        added_edges = 0
        disease_snps_dict = self.get_disease_snps_dict()
        hgnc_id_gene_rid_cache = {}

        for disease, rows in disease_snps_dict.items():

            for snp in tqdm(rows, desc=f'Add has_X_snp_cv edges to BEL for {disease}'):
                if snp.hgnc_id in hgnc_id_gene_rid_cache:
                    gene_mapped_rid = hgnc_id_gene_rid_cache[snp.hgnc_id]
                else:
                    gene_mapped_rid = self._get_set_gene_rid(hgnc_id=snp.hgnc_id)
                    hgnc_id_gene_rid_cache[snp.hgnc_id] = gene_mapped_rid

                if gene_mapped_rid:
                    snp_rid = self.get_create_rid('snp', {'rs_number': "rs" + str(snp.rs_number)})
                    value_dict = {'clinical_significance': snp.clinical_significance,
                                  'phenotype': snp.phenotype,
                                  'keyword': snp.keyword}
                    self.create_edge(class_name='has_mapped_snp_cv',
                                     from_rid=gene_mapped_rid,
                                     to_rid=snp_rid,
                                     value_dict=value_dict,
                                     if_not_exists=True)
                    added_edges += 1

                    # fetch all down and upstream gene_rids
                    gene_type_rids = self.get_set_gene_rids_by_position(
                        chromosome=snp.chromosome,
                        position=snp.position,
                        gene_types=['downstream', 'upstream']
                    )

                    for gene_type, gene_rids in gene_type_rids.items():
                        for gene_rid in gene_rids:
                            class_name = f"has_{snp_type[gene_type]}_snp_cv"
                            self.create_edge(class_name=class_name,
                                             from_rid=gene_rid,
                                             to_rid=snp_rid,
                                             value_dict=value_dict)
                            added_edges += 1
        return added_edges

    def _get_set_gene_rid(self, hgnc_id: str):
        """Insert gene (if not exists) and returns OrientDB @rid."""
        gene_rid = None

        hgnc_rids = self.query_class('hgnc', columns=['symbol'], limit=1, id=hgnc_id)

        if hgnc_rids:
            gene = hgnc_rids[0]
            bel = f'g(HGNC:"{gene["symbol"]}")'
            data = {'pure': True,
                    'bel': bel,
                    'name': gene['symbol'],
                    'namespace': "HGNC"
                    }
            gene_rid = self.get_create_rid('gene', data, check_for='bel')

        return gene_rid



"""DisGeNet RDBMS model definition."""
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Float, BigInteger, ForeignKey

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class DisgenetGene(Base):
    """Class definition for the disgenet_gene table."""

    __tablename__ = 'disgenet_gene'
    id = Column(Integer, primary_key=True)

    gene_id = Column(Integer, ForeignKey('disgenet_gene_symbol.gene_id'))
    gene_symbol = relationship("DisgenetGeneSymbol", back_populates='gene_disease_pmid_associations')
    disease_id = Column(String(100), ForeignKey('disgenet_disease.disease_id'))
    disease = relationship("DisgenetDisease", foreign_keys=[disease_id])
    score = Column(Float)
    pmid = Column(BigInteger)
    source_id = Column(Integer, ForeignKey('disgenet_source.id'))
    source = relationship("DisgenetSource", foreign_keys=[source_id])

    def as_dict(self):
        """Convert object values to dictionary."""
        rs = object_as_dict(self, exclude=['id', 'source_id'])
        rs.update({
            'gene_symbol': self.gene_symbol.gene_symbol,
            'disease_name': self.disease.disease_name,
            'source': self.source.source
        })
        return rs


class DisgenetGeneSymbol(Base):
    """Class definition for the disgenet_gene_symbol table."""

    __tablename__ = 'disgenet_gene_symbol'
    gene_id = Column(Integer, primary_key=True)
    gene_symbol = Column(String(50), index=True)

    gene_disease_pmid_associations = relationship("DisgenetGene", back_populates="gene_symbol")

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)


class DisgenetVariant(Base):
    """Class definition for the disgenet_variant table."""

    __tablename__ = 'disgenet_variant'
    id = Column(Integer, primary_key=True)

    snp_id = Column(String(20), index=True)
    chromosome = Column(String(2))
    position = Column(BigInteger)
    disease_id = Column(String(100), ForeignKey('disgenet_disease.disease_id'))
    disease = relationship("DisgenetDisease", foreign_keys=[disease_id])
    score = Column(Float)
    pmid = Column(BigInteger, index=True)
    source_id = Column(Integer, ForeignKey('disgenet_source.id'))
    source = relationship("DisgenetSource", foreign_keys=[source_id])

    def as_dict(self):
        """Convert object values to dictionary."""
        rs = object_as_dict(self, exclude=['id', 'source_id'])
        rs.update({
            'disease_name': self.disease.disease_name,
            'source': self.source.source
        })
        return rs


class DisgenetDisease(Base):
    """Class definition for the disgenet_disease table."""

    __tablename__ = 'disgenet_disease'
    disease_id = Column(String(100), primary_key=True)
    disease_name = Column(String(255), index=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)


class DisgenetSource(Base):
    """Class definition for the disgenet_source table."""

    __tablename__ = 'disgenet_source'
    id = Column(Integer, primary_key=True)
    source = Column(String(100), index=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self, exclude=['id'])



"""Root init for eBEL."""
from . import parser
from . import constants
from . import errors
from . import transformers
from . import cache

from .manager.orientdb.biodbs.bel import Bel

__version__ = '1.0.5'

__title__ = 'e(BE:L)'
__description__ = 'Validation and extension of biomedical knowledge graphs'
__url__ = 'https://github.com/e-bel/ebel'

__author__ = 'Christian Ebeling'
__email__ = 'christian.ebeling@scai.fraunhofer.de'

__license__ = '?'
__copyright__ = """Copyright (c) 2021 Christian Ebeling, Fraunhofer Institute for Algorithms and Scientific
Computing SCAI, Schloss Birlinghoven, 53754 Sankt Augustin, Germany"""

project_name = __title__



"""DrugBank RDBMS model definition."""

from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Text, ForeignKey, Date

Base = declarative_base()


class Drugbank(Base):
    """Class definition for the drugbank table."""

    __tablename__ = 'drugbank'
    id = Column(Integer, primary_key=True)
    drugbank_id = Column(String(10), index=True)
    name = Column(String(255))
    description = Column(Text)
    cas_number = Column(String(20))
    unii = Column(String(20))
    state = Column(String(20))
    indication = Column(Text)
    pharmacodynamics = Column(Text)
    toxicity = Column(Text)
    metabolism = Column(Text)
    absorption = Column(Text)
    half_life = Column(Text)
    route_of_elimination = Column(Text)
    volume_of_distribution = Column(Text)
    clearance = Column(Text)
    mechanism_of_action = Column(Text)
    fda_label = Column(Text)

    references = relationship("Reference", back_populates="drugbank", cascade="save-update")
    synonyms = relationship("Synonym", back_populates="drugbank", cascade="save-update")
    targets = relationship("Target", back_populates="drugbank", cascade="save-update")
    external_identifiers = relationship("ExternalIdentifier", back_populates="drugbank", cascade="save-update")
    product_names = relationship("ProductName", back_populates="drugbank", cascade="save-update")
    drug_interactions = relationship("DrugInteraction", back_populates="drugbank", cascade="save-update")
    statuses = relationship("Status", back_populates="drugbank", cascade="save-update")
    patents = relationship("Patent", back_populates="drugbank", cascade="save-update")
    pathways = relationship("Pathway", back_populates="drugbank", cascade="save-update")

    def __str__(self):
        """Class string definition."""
        return self.drugbank_id

    # TODO: add drug_interaction
    def as_dict(self):
        """Convert object values to dictionary."""
        return {'drugbank_id': self.drugbank_id,
                'name': self.name,
                'description': self.description,
                'cas_number': self.cas_number,
                'toxicity': self.toxicity,
                'indication': self.indication,
                'pharmacodynamics': self.pharmacodynamics,
                'metabolism': self.metabolism,
                'absorption': self.absorption,
                'half_life': self.half_life,
                'route_of_elimination': self.route_of_elimination,
                'volume_of_distribution': self.volume_of_distribution,
                'clearance': self.clearance,
                'mechanism_of_action': self.mechanism_of_action,
                'fda_label': self.fda_label,
                'references': [x.pmid for x in self.references],
                'pathways': [x.smpdb_id for x in self.pathways],
                'patents': [x.as_dict() for x in self.patents],
                'targets': [x.as_dict() for x in self.targets],
                'product_names': [x.name for x in self.product_names],
                'external_identifiers': [x.as_dict() for x in self.external_identifiers],
                'statuses': [x.status for x in self.statuses],
                }


class Pathway(Base):
    """Class definition for the drugbank_pathway table."""

    __tablename__ = 'drugbank_pathway'
    id = Column(Integer, primary_key=True)
    smpdb_id = Column(String(255))

    drugbank_id = Column(Integer, ForeignKey('drugbank.id'))
    drugbank = relationship("Drugbank", back_populates="pathways")

    def __str__(self):
        return self.smpdb_id

    def as_dict(self):
        """Convert object values to dictionary."""
        return {'smpdb_id': self.smpdb_id, 'drugbank_id': self.drugbank.drugbank_id}


class Patent(Base):
    """Class definition for the drugbank_patent table."""

    __tablename__ = 'drugbank_patent'
    id = Column(Integer, primary_key=True)
    number = Column(String(255))
    country = Column(String(255))
    approved = Column(Date)
    expires = Column(Date)
    pediatric_extension = Column(String(255))

    drugbank_id = Column(Integer, ForeignKey('drugbank.id'))
    drugbank = relationship("Drugbank", back_populates="patents")

    def __str__(self):
        return self.number

    def as_dict(self):
        """Convert object values to dictionary."""
        return {'number': self.number,
                'country': self.country,
                'approved': self.approved.strftime("%Y-%m-%d"),
                'expires': self.expires.strftime("%Y-%m-%d"),
                'pediatric_extension': self.pediatric_extension,
                'drugbank_id': self.drugbank.drugbank_id
                }


class Status(Base):
    """Class definition for the drugbank_status table."""

    __tablename__ = 'drugbank_status'
    id = Column(Integer, primary_key=True)
    status = Column(String(20), index=True)

    drugbank_id = Column(Integer, ForeignKey('drugbank.id'))
    drugbank = relationship("Drugbank", back_populates="statuses")

    def __str__(self):
        return self.status

    def as_dict(self):
        """Convert object values to dictionary."""
        return {'smpdb_id': self.status, 'drugbank_id': self.drugbank.drugbank_id}


class ExternalIdentifier(Base):
    """Class definition for the drugbank_external_identifier table."""

    __tablename__ = 'drugbank_external_identifier'
    id = Column(Integer, primary_key=True)
    resource = Column(String(255))
    identifier = Column(String(255), index=True)

    drugbank_id = Column(Integer, ForeignKey('drugbank.id'))
    drugbank = relationship("Drugbank", back_populates="external_identifiers")

    def __str__(self):
        return self.identifier

    def as_dict(self):
        """Convert object values to dictionary."""
        return {'resource': self.resource, 'identifier': self.identifier, 'drugbank_id': self.drugbank.drugbank_id}


class Reference(Base):
    """Class definition for the drugbank_reference table."""

    __tablename__ = 'drugbank_reference'
    id = Column(Integer, primary_key=True)
    pmid = Column(Integer)

    drugbank_id = Column(Integer, ForeignKey('drugbank.id'))
    drugbank = relationship("Drugbank", back_populates="references")

    def __str__(self):
        return self.pmid

    def as_dict(self):
        """Convert object values to dictionary."""
        return {'pmid': self.pmid, 'drugbank_id': self.drugbank.drugbank_id}


class Target(Base):
    """Class definition for the drugbank_target table."""

    __tablename__ = 'drugbank_target'
    id = Column(Integer, primary_key=True)
    uniprot = Column(String(20), index=True)
    action = Column(String(50), index=True)
    known_action = Column(String(20), index=True)

    drugbank_id = Column(Integer, ForeignKey('drugbank.id'))
    drugbank = relationship("Drugbank", back_populates="targets")

    def __str__(self):
        return self.uniprot

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'uniprot': self.uniprot,
            'action': self.action,
            'known_action': self.known_action,
            'drugbank_id': self.drugbank.drugbank_id
        }


class DrugInteraction(Base):
    """Class definition for the drugbank_drug_interaction table."""

    __tablename__ = 'drugbank_drug_interaction'
    id = Column(Integer, primary_key=True)
    drugbank_id = Column(String(10), index=True)
    name = Column(Text)
    description = Column(Text)

    db_id = Column(Integer, ForeignKey('drugbank.id'))  # exception because drugbank_id is already a field
    drugbank = relationship("Drugbank", back_populates="drug_interactions")

    def __str__(self):
        return self.drugbank_id

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'drugbank_id': self.drugbank_id,
            'name': self.name,
            'description': self.description,
            'interactor_drugbank_id': self.drugbank.drugbank_id
        }


class ProductName(Base):
    """Class definition for the drugbank_product_name table."""

    __tablename__ = 'drugbank_product_name'
    id = Column(Integer, primary_key=True)
    name = Column(Text)

    drugbank_id = Column(Integer, ForeignKey('drugbank.id'))
    drugbank = relationship("Drugbank", back_populates="product_names")

    def __str__(self):
        return self.name

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'drugbank_id': self.drugbank.drugbank_id,
            'name': self.name
        }


class Synonym(Base):
    """Class definition for the drugbank_synonym table."""

    __tablename__ = 'drugbank_synonym'
    id = Column(Integer, primary_key=True)
    synonym = Column(Text)

    drugbank_id = Column(Integer, ForeignKey('drugbank.id'))
    drugbank = relationship("Drugbank", back_populates="synonyms")

    def __str__(self):
        return self.synonym

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'drugbank_id': self.drugbank.drugbank_id,
            'synonym': self.synonym
        }



"""Protein Atlas RDBMS model definition."""
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Text, Numeric

Base = declarative_base()


class ProteinAtlasNormalTissue(Base):
    """Class definition for the protein_atlas_normal_tissue table."""

    __tablename__ = 'protein_atlas_normal_tissue'
    id = Column(Integer, primary_key=True)

    gene = Column(String(100), index=True)
    gene_name = Column(String(100))
    tissue = Column(String(100))
    cell_type = Column(String(100))
    level = Column(String(100), index=True)
    reliability = Column(String(100), index=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return {'gene': self.gene,
                'gene_name': self.gene_name,
                'tissue': self.tissue,
                'cell_type': self.cell_type,
                'level': self.level,
                'reliability': self.reliability}


class ProteinAtlasSubcellularLocation(Base):
    """Class definition for the protein_atlas_subcellular_location table."""

    __tablename__ = 'protein_atlas_subcellular_location'
    id = Column(Integer, primary_key=True)

    gene = Column(String(100))
    gene_name = Column(String(100))
    reliability = Column(String(100))
    main_location = Column(String(100))
    additional_location = Column(String(100))
    extracellular_location = Column(String(100))
    enhanced = Column(String(100))
    supported = Column(String(100))
    approved = Column(String(100))
    uncertain = Column(String(100))
    single_cell_variation_intensity = Column(String(100))
    single_cell_variation_spatial = Column(String(100))
    cell_cycle_dependency = Column(Text)
    go_id = Column(Text)

    def as_dict(self):
        """Convert object values to dictionary."""
        return {'gene': self.gene,
                'gene_name': self.gene_name,
                'reliability': self.reliability,
                'main_location': self.main_location,
                'additional_location': self.additional_location,
                'extracellular_location': self.extracellular_location,
                'enhanced': self.enhanced,
                'supported': self.supported,
                'approved': self.approved,
                'uncertain': self.uncertain,
                'single_cell_variation_intensity': self.single_cell_variation_intensity,
                'single_cell_variation_spatial': self.single_cell_variation_spatial,
                'cell_cycle_dependency': self.cell_cycle_dependency,
                'go_id': self.go_id}


class ProteinAtlasRnaTissueConsensus(Base):
    """Class definition for the protein_atlas_rna_tissue_consensus table."""

    __tablename__ = 'protein_atlas_rna_tissue_consensus'
    id = Column(Integer, primary_key=True)

    gene = Column(String(100), index=True)
    gene_name = Column(String(100), index=True)
    tissue = Column(String(100), index=True)
    nx = Column(Numeric(8, 1))

    def as_dict(self):
        """Convert object values to dictionary."""
        return {'gene': self.gene,
                'gene_name': self.gene_name,
                'tissue': self.tissue,
                'nx': self.nx}


class ProteinAtlasRnaBrainGtex(Base):
    """Class definition for the protein_atlas_rna_brain_gtex table."""

    __tablename__ = 'protein_atlas_rna_brain_gtex'
    id = Column(Integer, primary_key=True)

    gene = Column(String(100), index=True)
    gene_name = Column(String(100), index=True)
    brain_region = Column(String(100), index=True)
    tpm = Column(Numeric(8, 1))
    p_tpm = Column(Numeric(8, 1))
    nx = Column(Numeric(8, 1))

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'gene': self.gene,
            'gene_name': self.gene_name,
            'brain_region': self.brain_region,
            'tpm': self.tpm,
            'p_tpm': self.p_tpm,
            'nx': self.nx
        }


class ProteinAtlasRnaBrainFantom(Base):
    """Class definition for the protein_atlas_rna_brain_fantom table."""

    __tablename__ = 'protein_atlas_rna_brain_fantom'
    id = Column(Integer, primary_key=True)

    gene = Column(String(100))
    gene_name = Column(String(100))
    brain_region = Column(String(100))
    tags_per_million = Column(String(100))
    scaled_tags_per_million = Column(String(100))
    nx = Column(String(100))

    def as_dict(self):
        """Convert object values to dictionary."""
        return {'gene': self.gene,
                'gene_name': self.gene_name,
                'brain_region': self.brain_region,
                'tags_per_million': self.tags_per_million,
                'scaled_tags_per_million': self.scaled_tags_per_million,
                'nx': self.nx}


class ProteinAtlasRnaMouseBrainAllen(Base):
    """Class definition for the protein_atlas_rna_mouse_brain_allen table."""

    __tablename__ = 'protein_atlas_rna_mouse_brain_allen'
    id = Column(Integer, primary_key=True)

    gene = Column(String(100))
    gene_name = Column(String(100))
    brain_region = Column(String(100))
    expression_energy = Column(Numeric(8, 1))

    def as_dict(self):
        """Convert object values to dictionary."""
        return {'gene': self.gene,
                'gene_name': self.gene_name,
                'brain_region': self.brain_region,
                'expression_energy': self.expression_energy}



"""Chebi."""
import logging
import re
from typing import Dict

import pandas as pd
from pyorient import OrientDB
from tqdm import tqdm

from ebel import tools
from ebel.manager.orientdb import odb_meta, urls
from ebel.manager.orientdb.constants import CHEBI
from ebel.manager.rdbms.models import chebi

logger = logging.getLogger(__name__)


class Chebi(odb_meta.Graph):
    """Chebi."""

    def __init__(self, client: OrientDB = None):
        """Init CHEBI."""
        self.client = client
        self.biodb_name = CHEBI
        self.urls = {chebi.Compound.__tablename__: urls.CHEBI_COMPOUND,
                     chebi.Relation.__tablename__: urls.CHEBI_RELATION,
                     chebi.Inchi.__tablename__: urls.CHEBI_INCHI,
                     chebi.ChemicalData.__tablename__: urls.CHEBI_CHEMICALDATA,
                     chebi.Comment.__tablename__: urls.CHEBI_COMMENT,
                     chebi.DatabaseAccession.__tablename__: urls.CHEBI_DATABASEACCESSION,
                     chebi.Name.__tablename__: urls.CHEBI_NAME,
                     chebi.Reference.__tablename__: urls.CHEBI_REFERENCE,
                     chebi.Structure.__tablename__: urls.CHEBI_STRUCTURE}

        super().__init__(urls=self.urls,
                         biodb_name=self.biodb_name,
                         tables_base=chebi.Base)

    def __len__(self) -> int:
        """Get number of edges in OrientDB."""
        sql = "Select count(*) from bel where namespace='CHEBI'"
        return self.execute(sql)[0].oRecordData['count']

    def __contains__(self, name: str) -> bool:
        """Checks if RS number (without prefix RS) exists in BEL graph."""
        sql = f'Select 1 from bel where namespace="CHEBI" and name = "{name}"'
        return bool(self.execute(sql))

    def insert_data(self) -> Dict[str, int]:
        """Insert data in generic OrientDB class."""
        logger.info(f'Insert data in {self.biodb_name}')

        inserted = {}

        file_path_compound = tools.get_file_path(urls.CHEBI_COMPOUND, self.biodb_name)
        df_compounds = pd.read_csv(file_path_compound, sep="\t", low_memory=False, error_bad_lines=False)
        df_compound_ids = df_compounds[['ID']]
        del df_compounds

        for table_name, url in tqdm(self.urls.items(), desc=f'Import {self.biodb_name.upper()}'):

            file_path = tools.get_file_path(url, self.biodb_name)
            seperator = "\t" if re.search(r'.*\.tsv(\.gz)?$', file_path) else ","
            encoding = "ISO-8859-1" if table_name == 'chebi_reference' else None

            dfs = pd.read_csv(file_path, sep=seperator, encoding=encoding, low_memory=False, error_bad_lines=False,
                              chunksize=100000)
            inserted[table_name] = 0
            for df in tqdm(dfs, desc=f'Import {table_name}'):
                df.columns = df.columns.str.lower()

                if table_name == chebi.Inchi.__tablename__:
                    # with_index = True
                    df.index += 1
                    df.index.rename('id', inplace=True)
                    df.rename(columns={'chebi_id': "compound_id"}, inplace=True)

                if 'compound_id' in df.columns:
                    df = df_compound_ids.rename(columns={'ID': 'compound_id'}).set_index('compound_id').join(
                        df.set_index('compound_id'), how='inner').reset_index()

                if 'init_id' in df.columns:
                    df = df_compound_ids.rename(columns={'ID': 'init_id'}).set_index('init_id').join(
                        df.set_index('init_id'), how='inner').reset_index()

                if 'final_id' in df.columns:
                    df = df_compound_ids.rename(columns={'ID': 'final_id'}).set_index('final_id').join(
                        df.set_index('final_id'), how='inner').reset_index()

                df.to_sql(table_name, self.engine, index=False, if_exists='append')

                inserted[table_name] += df.shape[0]
        self.session.commit()
        return inserted

    def update_bel(self) -> int:
        """Update the BEL edges with CHEBI metadata."""
        updated = 0
        sql = "Update {rid} set chebi = {chebi_id}"
        chebi_nodes = self.query_class(class_name='bio_object',
                                       columns=['name', '@class'],
                                       namespace='CHEBI',
                                       pure=True)

        for chebi_node in tqdm(chebi_nodes, desc='Update ChEBI identifier in BEL'):
            chebi_compound = self.session.query(chebi.Compound.id) \
                .filter(chebi.Compound.name == chebi_node['name']).first()

            if chebi_compound:
                updated += self.execute(sql.format(rid=chebi_node['rid'], chebi_id=chebi_compound[0]))[0]

        return updated

    def update_interactions(self) -> int:
        """Abstract method."""
        return 0



"""String constants for the OrientDB submodule."""


# Databases
BIOGRID = 'biogrid'
CHEBI = 'chebi'
CLINICAL_TRIALS = 'clinical_trials'
CLINVAR = 'clinvar'
DISGENET = 'disgenet'
DRUGBANK = 'drugbank'
ENSEMBL = 'ensembl'
FLYBASE = 'flybase'
GWAS_CATALOG = 'gwas_catalog'
HGNC = 'hgnc'
EXPRESSION_ATLAS = 'expression_atlas'
INTACT = 'intact'
IUPHAR = 'iuphar'
KEGG = 'kegg'
MGI = 'mgi'
MIRTARBASE = 'mirtarbase'
NCBI = 'ncbi'
OFFSIDES = 'offsides'
PATHWAY_COMMONS = 'pathway_commons'
PROTEIN_ATLAS = 'protein_atlas'
REACTOME = 'reactome'
RGD = 'rgd'
SIDER = 'sider'
STITCH = 'stitch'
STRINGDB = 'stringdb'
UNIPROT = 'uniprot'

# Data Types
NODES = 'nodes'
EDGES = 'edges'



"""UniProt RDBMS model definition."""
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Table, Text, ForeignKey

from collections import defaultdict

Base = declarative_base()

uniprot__uniprot_keyword = Table('uniprot__uniprot_keyword', Base.metadata,
                                 Column('uniprot_id', Integer, ForeignKey('uniprot.id')),
                                 Column('uniprot_keyword_id', Integer,
                                        ForeignKey('uniprot_keyword.keywordid'))
                                 )

uniprot__uniprot_host = Table('uniprot__uniprot_host', Base.metadata,
                              Column('uniprot_id', Integer, ForeignKey('uniprot.id')),
                              Column('uniprot_organism_id', Integer, ForeignKey('uniprot_organism.taxid'))
                              )

uniprot__uniprot_xref = Table('uniprot__uniprot_xref', Base.metadata,
                              Column('uniprot_id', Integer, ForeignKey('uniprot.id')),
                              Column('uniprot_xref_id', Integer, ForeignKey('uniprot_xref.id'))
                              )

uniprot__uniprot_subcellular_location = Table('uniprot__uniprot_subcellular_location', Base.metadata,
                                              Column('uniprot_id', Integer, ForeignKey('uniprot.id')),
                                              Column('uniprot_subcellular_location_id',
                                                     Integer, ForeignKey('uniprot_subcellular_location.id'))
                                              )


class Uniprot(Base):
    """Class definition for the UniProt table."""

    __tablename__ = 'uniprot'

    id = Column(Integer, primary_key=True)

    accession = Column(String(20), unique=True)
    name = Column(String(100), nullable=False, unique=True)
    recommended_name = Column(String(255), nullable=True)

    taxid = Column(Integer, ForeignKey('uniprot_organism.taxid'), nullable=False, index=True)
    organism = relationship("Organism")

    function_id = Column(Integer, ForeignKey('uniprot_function.id'), nullable=True)
    function = relationship("Function")

    gene_names = relationship("Gene", back_populates="uniprot")

    gene_symbol = relationship("GeneSymbol", uselist=False, back_populates="uniprot")

    keywords = relationship(
        "Keyword",
        secondary=uniprot__uniprot_keyword,
        back_populates="uniprots")

    hosts = relationship(
        "Organism",
        secondary=uniprot__uniprot_host,
        back_populates="uniprots"
    )

    xrefs = relationship(
        "Xref",
        secondary=uniprot__uniprot_xref,
        back_populates="uniprots"
    )

    subcellular_locations = relationship(
        "SubcellularLocation",
        secondary=uniprot__uniprot_subcellular_location,
        back_populates="uniprots"
    )

    def __repr__(self):
        return self.name

    def as_dict(self):
        """Convert object values to dictionary."""
        xrefs_grouped = defaultdict(list)
        for xref in self.xrefs:
            xrefs_grouped[xref.db].append(xref.identifier)
        xrefs_grouped = {k: sorted(v) for k, v in xrefs_grouped.items()}

        return {
            'name': self.name,
            'accession': self.accession,
            'recommended_name': self.recommended_name,
            'taxid': self.taxid,
            'function_description': self.function.description if self.function else self.function,
            'gene_names': [x.name for x in self.gene_names],
            'gene_symbol': self.gene_symbol.symbol if self.gene_symbol else self.gene_symbol,
            'keywords': [{'keyword': x.keyword_name, 'id': x.keywordid} for x in self.keywords],
            'hosts': [{'name': x.scientific_name, 'taxid': x.taxid} for x in self.hosts],
            'xrefs': xrefs_grouped,
            'subcellular_locations': [x.name for x in self.subcellular_locations],
            'organism': self.organism.scientific_name
        }


class GeneSymbol(Base):
    """Class definition for the uniprot_gene_symbol table."""

    __tablename__ = 'uniprot_gene_symbol'
    id = Column(Integer, primary_key=True)
    symbol = Column(String(100), nullable=False, index=True)
    uniprot_id = Column(Integer, ForeignKey('uniprot.id'))
    uniprot = relationship("Uniprot", back_populates="gene_symbol")

    def __repr__(self):
        """Define repr."""
        return self.symbol


class Gene(Base):
    """Class definition for the uniprot_gene table."""

    __tablename__ = 'uniprot_gene'
    id = Column(Integer, primary_key=True)
    name = Column(String(100), nullable=False, index=True)
    uniprot_id = Column(Integer, ForeignKey('uniprot.id'))
    uniprot = relationship("Uniprot", back_populates="gene_names")


class Keyword(Base):
    """Class definition for the uniprot_keyword table."""

    __tablename__ = 'uniprot_keyword'

    keywordid = Column(Integer, primary_key=True)
    keyword_name = Column(String(100), index=True)

    uniprots = relationship(
        "Uniprot",
        secondary=uniprot__uniprot_keyword,
        back_populates="keywords")

    def __repr__(self):
        """Define repr."""
        return f"{self.keyword_name}[{self.keywordid}]"


class Organism(Base):
    """Class definition for the uniprot_organism table."""

    __tablename__ = 'uniprot_organism'

    taxid = Column(Integer, primary_key=True)
    scientific_name = Column(String(255))  # TODO:Check if index=True with  is possible

    uniprots = relationship(
        "Uniprot",
        secondary=uniprot__uniprot_host,
        back_populates="hosts")


class SubcellularLocation(Base):
    """Class definition for the uniprot_subcellular_location table."""

    __tablename__ = 'uniprot_subcellular_location'

    id = Column(Integer, primary_key=True)

    name = Column(String(100), index=True)

    uniprots = relationship(
        "Uniprot",
        secondary=uniprot__uniprot_subcellular_location,
        back_populates="subcellular_locations")


class Xref(Base):
    """Class definition for the uniprot_xref table."""

    __tablename__ = 'uniprot_xref'

    id = Column(Integer, primary_key=True)

    db = Column(String(50), index=True)
    identifier = Column(String(100), index=True)

    uniprots = relationship(
        "Uniprot",
        secondary=uniprot__uniprot_xref,
        back_populates="xrefs")


class Function(Base):
    """Class definition for the uniprot_function table."""

    __tablename__ = 'uniprot_function'

    id = Column(Integer, primary_key=True)

    description = Column(Text)

    uniprots = relationship("Uniprot", back_populates="function")



"""Variables used by all tests in the test_manager directory."""
import os
import pathlib

from ebel import Bel
from ebel.constants import DEFAULT_ODB
from ebel.tools import get_config_value
from ebel.manager.orientdb.odb_meta import Graph


# Paths
VALIDATION_TEST_DIR = pathlib.Path(__file__).parent.absolute()
TEST_DATA_DIR = os.path.join(VALIDATION_TEST_DIR, "..", "data")

IMPORT_DIR = os.path.join(TEST_DATA_DIR, "import_tests")
TEST_BEL = os.path.join(IMPORT_DIR, "basic_import_test.bel")
TEST_JSON = TEST_BEL + ".json"

# Parameters for TEST database
USER = get_config_value(DEFAULT_ODB, 'user')
PASSWORD = get_config_value(DEFAULT_ODB, 'password')
DB_NAME = 'ebel_test'
SERVER = 'localhost'
PORT = "2424"
ROOT_PWD = get_config_value(DEFAULT_ODB, 'root_password')

if ROOT_PWD is None:
    raise ValueError("Need root password to perform tests. Please add 'root_password' to configuration file")

# Configure client
Graph.set_configuration(name=DB_NAME,
                        user=USER,
                        password=PASSWORD,
                        server=SERVER,
                        port=PORT,
                        root_password=ROOT_PWD)
test_client = Bel()



"""Defined variables used throughout the package."""
# -*- coding: utf-8 -*-

import os

THIS_DIR = os.path.dirname(__file__)
PROJECT_NAME = 'ebel'

HOME = os.path.expanduser('~')
LIBRARY_NAME = PROJECT_NAME

# Path to folder
PROJECT_DIR = os.path.join(HOME, f".{PROJECT_NAME}")
if not os.path.exists(PROJECT_DIR):
    os.mkdir(PROJECT_DIR)

# Path to data folder
DATA_DIR = os.path.join(PROJECT_DIR, 'data')
if not os.path.exists(DATA_DIR):
    os.mkdir(DATA_DIR)

# Path to logs folder
LOG_DIR = os.path.join(PROJECT_DIR, 'logs')
if not os.path.exists(LOG_DIR):
    os.mkdir(LOG_DIR)

# Default database name and location
DB_NAME = '{}.db'.format(PROJECT_NAME)
DB_PATH = os.path.join(DATA_DIR, DB_NAME)

GRAMMAR_BEL_PATH = {'2': os.path.join(THIS_DIR, 'grammar', 'grammar_bel_2.bnf'),
                    '2_1': os.path.join(THIS_DIR, 'grammar', 'grammar_bel_2_1.bnf'),
                    }

GRAMMAR_NS_ANNO_PATH = os.path.join(THIS_DIR, 'grammar', 'grammar_belns_belanno_1__2.bnf')
GRAMMAR_START_NS = 'belns'
GRAMMAR_START_ANNO = 'belanno'
GRAMMAR_START_LINE = 'script_line_by_line'

# Variables
RID = 'rid'

URL = "URL"
PATTERN = "PATTERN"
LIST = "LIST"
FILE = "FILE"
ALLOWED_TYPES = [URL, LIST, PATTERN, FILE]

DEFAULT_ODB = "DEFAULT_ODB"

# Protein Modifications
PMOD = {
    'pmod_ace': "Ac",
    'pmod_adr': "ADPRib",
    'pmod_add': "ADP-rybosylation",
    'pmod_far': "Farn",
    'pmod_ger': "Gerger",
    'pmod_gly': "Glyco",
    'pmod_hyd': "Hy",
    'pmod_isg': "ISG",
    'pmod_me0': "Me",
    'pmod_me1': "Me1",
    'pmod_mon': "monomethylation",
    'pmod_me2': "Me2",
    'pmod_me3': "Me3",
    'pmod_tri': "trimethylation",
    'pmod_myr': "Myr",
    'pmod_ned': "Nedd",
    'pmod_ngl': "NGlyco",
    'pmod_nit': "NO",
    'pmod_ogl': "OGlyco",
    'pmod_pal': "Palm",
    'pmod_pho': "Ph",
    'pmod_sul': "Sulf",
    'pmod_sup': "sulphation",
    'pmod_suh': "sulfonation",
    'pmod_sum': "Sumo",
    'pmod_suy': "Ub",
    'pmod_ubi': "ubiquitinylation",
    'pmod_u48': "UbK48",
    'pmod_u63': "UbK63",
    'pmod_ubm': "UbMono",
    'pmod_ubp': "UbPoly"
}

CYTOSCAPE_TEMPLATE = """{
  "format_version" : "1.0",
  "generated_by" : "cytoscape-3.6.0",
  "target_cytoscapejs_version" : "~2.1",
  "data" : {
    "shared_name" : "BEL_in_Cytoscape",
    "name" : "BEL_in_Cytoscape",
    "SUID" : 100000000,
    "__Annotations" : [ "" ],
    "networkMetadata" : "",
    "selected" : false
  },
  "elements" : {
    "nodes" : %s,
    "edges" : %s
  }
}"""

MD_FILE_HEADER = """Error report"""

SPECIES_NAMESPACE = {
    "HGNC": 9606,
    "MGI": 10090,
    "RGD": 10116,
    "FLYBASE": 7227,
}

# Edge types
# Causal
INCREASES = 'increases'
DIRECTLY_INCREASES = 'directly_increases'
CAUSAL_INCREASE = {INCREASES, DIRECTLY_INCREASES}

DECREASES = 'decreases'
DIRECTLY_DECREASES = 'directly_decreases'
CAUSAL_DECREASE = {DECREASES, DIRECTLY_DECREASES}

# Correlative
POSITIVE_CORRELATION = 'positive_correlation'
NEGATIVE_CORRELATION = 'negative_correlation'


# Terminal Colors
class TerminalFormatting:
    RESET = "\033[0m"

    class Format:
        BOLD = "\033[1m"
        ITALIC = "\033[3m"
        UNDERLINED = "\033[4m"
        STRIKETHROUGH = "\033[9m"

    class Fore:
        __temp = "\033[{}m"
        __temp_b = "\033[1;{}m"
        RED = __temp.format(31)
        RED_BRIGHT = __temp_b.format(31)
        GREEN = __temp.format(32)
        GREEN_BRIGHT = __temp_b.format(32)
        YELLOW = __temp.format(33)
        YELLOW_BRIGHT = __temp_b.format(33)
        BLUE = __temp.format(34)
        BLUE_BRIGHT = __temp_b.format(34)
        PINK = __temp.format(35)
        PINK_BRIGHT = __temp_b.format(35)
        CYAN = __temp.format(36)
        CYAN_BRIGHT = __temp_b.format(36)
        WHITE = __temp_b.format(37)
        BLACK = __temp.format(30)
        GREY = __temp_b.format(30)
        GREY_BRIGHT = __temp_b.format(37)

    class Back:
        __temp = "\033[{}m"
        NOTHING = __temp.format(40)
        RED = __temp.format(41)
        GREEN = __temp.format(42)
        YELLOW = __temp.format(43)
        BLUE = __temp.format(44)
        PINK = __temp.format(45)
        BLUE_LIGHT = __temp.format(46)
        GREY = __temp.format(47)

    TITLE = f"{Back.GREY}{Fore.BLUE}{Format.BOLD}{Format.UNDERLINED}"
    HEADER = f"{Fore.YELLOW}{Format.BOLD}{Format.UNDERLINED}"
    QUESTION = f"{Fore.GREEN}{Format.ITALIC}"
    DEFAULT_VALUE = f"{Fore.GREEN_BRIGHT}"




# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'e(BE:L)'
copyright = '2021, Christian Ebeling, Bruce Schultz'
author = 'Christian Ebeling, Bruce Schultz'

# The full version, including alpha/beta/rc tags
release = '1.0.5'

extensions = [
    'sphinx.ext.autodoc',  # Core Sphinx library for auto html doc generation from docstrings
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    "sphinx_autodoc_typehints",  # Automatically document param types (less noise in class signature)
    'sphinx.ext.autosummary',  # Create neat summary tables for modules/classes/methods etc
    'sphinx_click.ext',
    'numpydoc',  # Parses docstrings that us numpy style
    'sphinxcontrib.openapi',  # For swagger file
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

language = None
pygments_style = 'sphinx'
todo_include_todos = False

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}

numpydoc_show_class_members = False

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []



"""OFFSIDES module."""

import logging
import pandas as pd

from tqdm import tqdm
from typing import Dict
from pyorient import OrientDB

from ebel.tools import get_file_path
from ebel.constants import RID
from ebel.manager.orientdb.constants import OFFSIDES
from ebel.manager.orientdb import odb_meta, urls, odb_structure

from ebel.manager.rdbms.models import offsides

logger = logging.getLogger(__name__)


class Offsides(odb_meta.Graph):
    """Drug side effects and drug-drug interactions were mined from publicly available data.

    Offsides is a database
    of drug side-effects that were found, but are not listed on the official FDA label. Twosides is the only
    comprehensive database drug-drug-effect relationships. Over 3,300 drugs and 63,000 combinations connected to
    millions of potential adverse reactions.

    More information can be found on http://tatonettilab.org/offsides/
    """

    def __init__(self, client: OrientDB = None):
        """Init SIDER."""
        self.client = client
        self.biodb_name = OFFSIDES
        self.urls = {OFFSIDES: urls.OFFSIDES}
        super().__init__(edges=odb_structure.offsides_edges,
                         nodes=odb_structure.offsides_nodes,
                         tables_base=offsides.Base,
                         urls=self.urls,
                         biodb_name=self.biodb_name)

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    def insert_data(self) -> Dict[str, int]:
        """Insert data in generic OrientDB class.

        Columns:

        drug_rxnorn_id 	integer 	RxNorm CUI (RxCUI) of each drug
        drug_concept_name 	string 	String name of each drug
        condition_meddra_id 	integer 	MedDRA code of each (adverse event) condition
        condition_concept_name 	string 	String name of each condition
        A 	integer 	Number of reports prescribed the drug that had the condition
        B 	integer 	Number of reports prescribed the drug that did not have the condition
        C 	integer 	Number of reports not prescribed the drug† that had the condition
        D 	integer 	Number of reports not prescribed the drug† that did not have the condition
        PRR 	float 	Proportional reporting ratio*
        PRR_error 	float 	Proportional reporting ratio error*
        mean_reporting_frequency 	float 	a / (a + b)

        Further documentation see here:
        https://github.com/tatonetti-lab/nsides-release/blob/master/release-notes/v0.1.md
        """
        file_path = get_file_path(self.urls[OFFSIDES], self.biodb_name)
        df = pd.read_csv(file_path, low_memory=False)
        # Because of repeating header, we have to delete all rows which are equal to the columns
        df = df[df.ne(df.columns).any(1)]
        df.drop_duplicates(inplace=True)
        df.columns = self._standardize_column_names(df.columns)
        df.index += 1
        df.index.rename('id', inplace=True)
        df.to_sql(offsides.Offsides.__tablename__,
                  self.engine,
                  if_exists='append',
                  chunksize=10000)
        return {self.biodb_name: df.shape[0]}

    def update_bel(self) -> int:
        """Create has_side_effect edges between drugs (drugbank) and side_effects.

        RxCUI (https://rxnav.nlm.nih.gov/RxNormAPIs.html#) is used as the mapping identifier.
        """
        self.clear_edges()
        self.delete_nodes_with_no_edges('side_effect')
        self.delete_nodes_with_no_edges('drug')

        # TODO: Translate to sqlalchemy query
        sql_temp = """Select
            o.condition_meddra_id,
            o.condition_concept_name,
            o.prr,
            o.mean_reporting_frequency
        from
            drugbank as d inner join
            drugbank_external_identifier as dei on (d.id=dei.drugbank_id) inner join
            offsides as o on (dei.identifier=o.drug_rxnorn_id)
        where
            d.drugbank_id='{}' and resource='RxCUI' and mean_reporting_frequency>=0.01
        group by
            o.condition_meddra_id,
            o.condition_concept_name,
            o.prr,
            o.mean_reporting_frequency
        """

        drugbank_ids = self.query_class('drug', columns=['drugbank_id'])
        drugbank_id_rids = {d['drugbank_id']: d[RID] for d in drugbank_ids}

        side_effects = self.query_class('side_effect', columns=['condition_meddra_id'])
        side_effect_rids = {d['condition_meddra_id']: d[RID] for d in side_effects}

        updated = 0

        for drugbank_id, drugbank_rid in tqdm(drugbank_id_rids.items(), desc=f'Update {self.biodb_name.upper()}'):
            for r in self.engine.execute(sql_temp.format(drugbank_id)):
                (condition_meddra_id,
                 condition_concept_name,
                 prr,
                 mean_reporting_frequency) = r

                if condition_meddra_id not in side_effect_rids:
                    side_effect_rid = self.insert_record(
                        'side_effect',
                        {
                            'label': condition_concept_name,
                            'condition_meddra_id': condition_meddra_id
                        }
                    )
                    side_effect_rids[condition_meddra_id] = side_effect_rid

                value_dict = {
                    'prr': prr,
                    'mean_reporting_frequency': mean_reporting_frequency
                }
                side_effect_rid = side_effect_rids[condition_meddra_id]
                self.create_edge(class_name='has_side_effect',
                                 from_rid=drugbank_rid,
                                 to_rid=side_effect_rid,
                                 value_dict=value_dict)
                updated += 1
        return updated

    def update_interactions(self) -> int:
        """Update edges with OFFSIDES metadata."""
        # not implemented because update_bel is implemented
        pass



"""IUPHAR API methods."""
from flask.globals import request
from ebel.manager.orientdb.odb_structure import iuphar_edges

from ebel.web.api import RDBMS
from ebel.manager.rdbms.models import iuphar
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_interaction():
    """Get generic IUPHAR entry."""
    return _get_data(iuphar.IupharInteraction)


def get_ligandby_by_id():
    """Get IUPHAR ligand entry by ID."""
    ligand_id = request.args.get('id')
    return RDBMS.get_session().query(iuphar.IupharLigand).get(ligand_id).as_dict()


def get_interaction_by_target_uniprot():
    """Get IUPHAR interaction entry by target UniProt ID."""
    return get_interaction()


def get_interaction_by_target_gene_symbol():
    """Get IUPHAR interaction entry by target gene symbol."""
    return get_interaction()


def get_ebel_relation():
    """Get Iuphar related eBEL relations."""
    as_iuphar_edge = [x for x in iuphar_edges if x.name == 'iuphar_interaction'][0]
    conf = {x.prop_name: x.prop_name for x in as_iuphar_edge.props}
    conf.update({
        'edge_id': "@rid.asString()",
        'interactor_a_rid': "out.@rid.asString()",
        'interactor_a_name': "out.name",
        'interactor_a_namespace': "out.namespace",
        'interactor_b_rid': "in.@rid.asString()",
        'interactor_b_namespace': "in.namespace",
        'interactor_b_name': "in.name",
        'relation': "@class.asString()"
    })
    sql = "SELECT "
    sql += ', '.join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM iuphar_interaction"

    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}
    pmid = request.args.get('pmid')
    if paras or pmid:
        wheres = []
        if paras:
            wheres += [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]
        if pmid:
            wheres += [f"{pmid} in pmids"]

        sql += " WHERE " + ' AND '.join(wheres)

    return _get_paginated_ebel_query_result(sql)



"""True values to check against."""

# True values after importing basic_import_test.bel.json with only extension - test_import_json
NODES_EXTENSION = {
    'bel': 12,
    'nn': 12,
    'pure_object': 11,
    'location_object': 10,
    'bio_concept': 0,
    'biological_process': 0,
    'pathology': 0,
    'bio_object': 11,
    'complex': 0,
    'abundance': 1,
    'population': 0,
    'genetic_flow': 10,
    'gene': 4,
    'rna': 3,
    'protein': 3,
    'micro_rna': 0,
    'bio_act': 1,
    'activity': 1,
    'reaction': 0,
    'degradation': 0,
    'cell_secretion': 0,
    'translocation': 0,
    'cell_surface_expression': 0,
    'bio_list': 0,
    'list': 0,
    'composite': 0,
    'ebel': 0,
    'variant': 0,
    'fragment': 0,
    'location': 0,
    'pmod': 0,
    'gmod': 0,
    'from_location': 0,
    'to_location': 0,
    'reactants': 0,
    'products': 0,
    'fusion_protein': 0,
    'fusion_rna': 0,
    'fusion_gene': 0
}
EDGES_EXTENSION = {
    'bel_relation': 9,
    'causal': 3,
    'increases': 1,
    'directly_increases': 1,
    'decreases': 1,
    'directly_decreases': 0,
    'rate_limiting_step_of': 0,
    'causes_no_change': 0,
    'regulates': 0,
    'correlative': 0,
    'negative_correlation': 0,
    'positive_correlation': 0,
    'association': 0,
    'no_correlation': 0,
    'genomic': 6,
    'orthologous': 0,
    'transcribed_to': 3,
    'translated_to': 3,
    'other': 0,
    'has_member': 0,
    'has_members': 0,
    'has_component': 0,
    'has_components': 0,
    'equivalent_to': 0,
    'is_a': 0,
    'sub_process_of': 0,
    'deprecated': 0,
    'analogous_to': 0,
    'biomarker_for': 0,
    'prognostic_biomarker_for': 0,
    'compiler': 0,
    'acts_in': 0,
    'has_product': 0,
    'has_variant': 0,
    'has_modification': 0,
    'reactant_in': 0,
    'translocates': 0,
    'includes': 0,
    'ebel_relation': 1,
    'has__protein': 1,
    'has__rna': 0,
    'has__gene': 0,
    'has__abundance': 0,
    'has__population': 0,
    'has__location': 0,
    'has__from_location': 0,
    'has__to_location': 0,
    'has__fragment': 0,
    'has__pmod': 0,
    'has__gmod': 0,
    'has__complex': 0,
    'has__micro_rna': 0,
    'has__variant': 0,
    'has__reactants': 0,
    'has__products': 0,
    'has__composite': 0,
    'has_fragmented_protein': 0,
    'has_modified': 0,
    'has_modified_protein': 0,
    'has_modified_gene': 0,
    'has_variant_obj': 0,
    'has_variant_gene': 0,
    'has_variant_rna': 0,
    'has_variant_protein': 0,
    'has_variant_micro_rna': 0,
    'has_located': 0,
    'has_located_gene': 0,
    'has_located_rna': 0,
    'has_located_protein': 0,
    'has_located_micro_rna': 0,
    'has_located_complex': 0,
    'has_located_abundance': 0,
    'has_located_population': 0,
    'pathway_interaction': 0,
    'has_ppi': 0
}



"""Expression Atlas API methods."""
import math
import re
import json
from ebel.web.api import RDBMS
from . import _get_paginated_query_result, add_query_filters

from ebel.web.api.ebel.v1 import _get_data
from flask import request
from sqlalchemy import inspect

from collections import Counter
from ebel.manager.rdbms.models.expression_atlas import (
    GroupComparison,
    Gsea,
    FoldChange,
    Experiment,
    Idf,
    SdrfCondensed)
from ebel import Bel

models_dict = {
    'experiment': Experiment,
    'idf': Idf,
    'gsea': Gsea,
    'group_comparison': GroupComparison,
    'foldchange': FoldChange,
    'sdrf_condensed': SdrfCondensed
}


def get_expriment():
    """Get list of experiments."""
    return _get_data(Experiment)


def get_idf():
    """Get list of IDFs."""
    return _get_data(Idf)


def get_group_comparison():
    """Get list of group comparisons."""
    return _get_data(GroupComparison)


def get_fold_change():
    """Get list of fold changes."""
    return _get_data(FoldChange)


def get_sdrf_condensed():
    """Get list of condensed version of SDRFs."""
    return _get_data(SdrfCondensed)


def get_gsea():
    """Get GSEA data."""
    return _get_data(Gsea, order_by=[Gsea.p_adj_non_dir])


class GseaType:
    """String constant definitions."""

    GO = 'go'
    REACTOME = 'reactome'
    INTERPRO = 'interpro'


def get_most_common_gseas_by_group_comparison_ids() -> list:
    """Identify most often ocurring GSEAs based on group comparison IDs."""
    b = Bel()
    group_comparison_ids_str = request.args.get('group_comparison_ids', '')
    group_comparison_ids = [
        int(x.strip())
        for x in group_comparison_ids_str.split(',')
        if isinstance(x, str) and re.search(r"\s*\d+\s*", x)
    ]
    query = b.session.query(
        Gsea.accession,
        Gsea.term,
        Gsea.gsea_type
    ).filter(Gsea.group_comparison_id.in_(group_comparison_ids))

    gsea_type = request.args.get('gsea_type', '')
    gsea_type = gsea_type if gsea_type in [GseaType.GO, GseaType.REACTOME, GseaType.INTERPRO] else ''

    if gsea_type:
        query = query.filter(Gsea.gsea_type == gsea_type)

    p_adj_non_dir = request.args.get('p_adj_non_dir', '')
    p_adj_non_dir = float(p_adj_non_dir) if re.search(r'\s*\d+(\.\d+)?\s*', p_adj_non_dir) else None
    if isinstance(p_adj_non_dir, float):
        query = query.filter(Gsea.p_adj_non_dir <= p_adj_non_dir)

    results = [x for x in query.all()]

    return [{'accession': x[0][0], 'term': x[0][1], 'type': x[0][2], 'occurence': x[1]}
            for x in Counter(results).most_common() if x[1] > 1]


def get_comparison_groups_by_genes():
    """Get experiments, comparion groups and foldchanges by 2 genes."""
    gene_1 = request.args.get('gene_symbol_1')
    gene_2 = request.args.get('gene_symbol_2')
    return _get_comparison_groups_by_genes(gene_1, gene_2)


def _get_comparison_groups_by_genes(gene_1, gene_2):
    """Get experiments, comparion groups and foldchanges by 2 genes."""
    result_keys = [
        'fold_change_gene_symbol_1',
        'fold_change_gene_symbol_2',
        'group_comparison',
        'experiment_title',
        'experiment_identifier',
        'is_positive_correlated'
    ]
    sql = f"""Select 
        a.log2foldchange, 
        b.log2foldchange, 
        gc.name, 
        e.title, 
        e.name, 
        if(((a.log2foldchange>0 and b.log2foldchange>0) or (a.log2foldchange<0 and b.log2foldchange<0)),1,0) 
    from (Select 
            group_comparison_id, 
            log2foldchange 
        from 
            expression_atlas_foldchange where gene_name = '{gene_1}') a 
    inner join (Select 
            group_comparison_id, 
            log2foldchange 
        from expression_atlas_foldchange where gene_name = '{gene_2}') b 
    using(group_comparison_id) 
    inner join expression_atlas_group_comparison gc on (gc.id=group_comparison_id) 
    inner join expression_atlas_experiment e on (e.id = gc.experiment_id)
    order by (a.log2foldchange + b.log2foldchange) desc"""
    with Bel().engine.connect() as con:
        return list([dict(zip(result_keys, x)) for x in con.execute(sql).fetchall()])


def get_comparison_groups_by_edge_rid():
    rid = request.args.get('edge_rid')
    res = Bel().query_get_dict(f"""Select 
                                    in.name as name_in,
                                    in.namespace as ns_in,
                                    out.name as name_out,
                                    out.namespace as ns_out,
                                    in.bel as in_bel,
                                    out.bel as out_bel,
                                    @class.asString() as relation,
                                    citation,
                                    annotation,
                                    evidence
                                from {rid}""")
    if(res):
        has_all_cols = all([(x in res[0]) for x in ['name_in', 'ns_in', 'name_out', 'ns_out']])
        if has_all_cols:
            both_ns_hgnc = res[0]['ns_in'] == 'HGNC' and res[0]['ns_out'] == 'HGNC'
            not_the_same = res[0]['name_in'] != res[0]['name_out']
            if has_all_cols and both_ns_hgnc and not_the_same:
                gene_symbol_1 = res[0]['name_in']
                gene_symbol_2 = res[0]['name_out']
                comparisons = _get_comparison_groups_by_genes(gene_symbol_1, gene_symbol_2)
                return {
                    'gene_symbol_1': gene_symbol_1,
                    'gene_symbol_2': gene_symbol_2,
                    'citation': res[0].get('citation'),
                    'annotation': res[0].get('annotation'),
                    'evidence': res[0].get('evidence'),
                    'relation': res[0].get('relation'),
                    'comparisons': comparisons,
                    'bel_object': res[0].get('in_bel'),
                    'bel_subject': res[0].get('out_bel'),
                }
    return []


def get_expression_atlas():
    columns = [
        Experiment.name,
        GroupComparison.name,
        Experiment.title,
        GroupComparison.group_comparison,
        GroupComparison.id
    ]
    data = json.loads(request.data)
    b = Bel()
    query = b.session.query(Experiment).join(GroupComparison)
    for table, columns_params in [(k, v) for k, v in data.items() if k not in ('page', 'page_size')]:
        any_value = any([v['value'].strip() for k, v in columns_params.items()])
        if any_value:
            model = models_dict[table]

            if model not in (Experiment, GroupComparison):
                query = query.join(model)
            query = add_query_filters(query, columns_params, model)
    query = query.with_entities(*columns)
    query = query.group_by(Experiment.id, GroupComparison.id)
    limit = int(data['page_size']) if data.get('page_size') else 10
    page = int(data['page']) if data.get('page') else 1
    count = query.count()
    query = query.limit(limit).offset(limit * (page-1))

    print(query.statement.compile(compile_kwargs={"literal_binds": True}))
    column_names = [f"{x.parent.class_.__tablename__}.{x.name}" for x in [inspect(c) for c in columns]]
    return {
        'page': page,
        'pages': math.ceil(count/limit),
        'count': count,
        'page_size': limit,
        'column_names': column_names,
        'results': [tuple(x) for x in query.all()]
    }



"""StringDB API methods."""

from flask import request

from ebel.manager.rdbms.models.stringdb import StringDb, StringDbAction
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_stringdb():
    """Get generic StringDB entry."""
    return _get_data(StringDb)


def get_stringdb_action():
    """Get StringDB entry by action."""
    return _get_data(StringDbAction)


def get_ebel_relation():
    """Get StringDB related eBEL relations."""
    conf = {
        'interactor_a_rid': "out.@rid.asString()",
        'interactor_a_name': "out.name",
        'interactor_a_namespace': "out.namespace",
        'interactor_a_uniprot': "out.uniprot",
        'edge_id': "@rid.asString()",
        'relation': "@class.asString()",
        'interactor_b_rid': "in.@rid.asString()",
        'interactor_b_namespace': "in.namespace",
        'interactor_b_name': "in.name",
        'interactor_b_uniprot': "in.uniprot",
    }
    sql = "SELECT "
    sql += ', '.join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_action_st"

    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}

    if paras:
        wheres = [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]
        sql += " WHERE " + ' AND '.join(wheres)

    return _get_paginated_ebel_query_result(sql)
